from __future__ import annotations

import json
import math
from collections.abc import Callable, Iterable, Mapping, Sequence
from contextlib import contextmanager
from dataclasses import dataclass, field
from itertools import chain
from typing import Any, TypeVar, cast

from xdsl.dialects.builtin import (
    AffineMapAttr,
    AffineSetAttr,
    AnyFloat,
    AnyUnrankedMemRefType,
    AnyUnrankedTensorType,
    AnyVectorType,
    ArrayAttr,
    BFloat16Type,
    BoolAttr,
    BytesAttr,
    ComplexType,
    DenseArrayBase,
    DenseResourceAttr,
    DictionaryAttr,
    Float16Type,
    Float32Type,
    Float64Type,
    Float80Type,
    Float128Type,
    FloatAttr,
    FunctionType,
    IndexType,
    IntAttr,
    IntegerAttr,
    IntegerType,
    LocationAttr,
    MemRefType,
    NoneAttr,
    NoneType,
    OpaqueAttr,
    Signedness,
    StridedLayoutAttr,
    StringAttr,
    SymbolRefAttr,
    TensorType,
    UnitAttr,
    UnrankedMemRefType,
    UnrankedTensorType,
    UnregisteredAttr,
    UnregisteredOp,
    VectorType,
    i1,
)
from xdsl.ir import (
    Attribute,
    Block,
    BlockArgument,
    Data,
    OpaqueSyntaxAttribute,
    Operation,
    ParametrizedAttribute,
    Region,
    SpacedOpaqueSyntaxAttribute,
    SSAValue,
    TypeAttribute,
    TypedAttribute,
)
from xdsl.traits import IsolatedFromAbove, IsTerminator
from xdsl.utils.base_printer import BasePrinter
from xdsl.utils.bitwise_casts import (
    convert_f16_to_u16,
    convert_f32_to_u32,
    convert_f64_to_u64,
)
from xdsl.utils.diagnostic import Diagnostic
from xdsl.utils.mlir_lexer import MLIRLexer


@dataclass(eq=False, repr=False)
class Printer(BasePrinter):
    print_generic_format: bool = field(default=False)
    print_properties_as_attributes: bool = field(default=False)
    print_debuginfo: bool = field(default=False)
    diagnostic: Diagnostic = field(default_factory=Diagnostic)

    _ssa_values: dict[SSAValue, str] = field(default_factory=dict, init=False)
    """
    maps SSA Values to their "allocated" names
    """
    _ssa_names: list[dict[str, int]] = field(
        default_factory=lambda: [dict[str, int]()], init=False
    )
    _block_names: list[dict[Block, int]] = field(
        default_factory=lambda: [dict[Block, int]()], init=False
    )
    _next_valid_name_id: list[int] = field(default_factory=lambda: [0], init=False)
    _next_valid_block_id: list[int] = field(default_factory=lambda: [0], init=False)

    @property
    def ssa_names(self):
        return self._ssa_names[-1]

    @property
    def block_names(self):
        return self._block_names[-1]

    @contextmanager
    def in_angle_brackets(self):
        self.print_string("<")
        try:
            yield
        finally:
            self.print_string(">")

    def print(self, *argv: Any) -> None:
        for arg in argv:
            if isinstance(arg, str):
                self.print_string(arg)
                continue
            if isinstance(arg, SSAValue):
                self.print_ssa_value(arg)
                continue
            if isinstance(arg, Attribute):
                self.print_attribute(arg)
                continue
            if isinstance(arg, Region):
                self.print_region(arg)
                continue
            if isinstance(arg, Block):
                self.print_block(arg)
                self._print_new_line()
                continue
            if isinstance(arg, Operation):
                self.print_op(arg)
                self._print_new_line()
                continue

            text = str(arg)
            self.print_string(text)

    K = TypeVar("K")
    V = TypeVar("V")

    def print_dictionary(
        self,
        elems: dict[K, V],
        print_key: Callable[[K], None],
        print_value: Callable[[V], None],
        delimiter: str = ", ",
    ) -> None:
        for i, (key, value) in enumerate(elems.items()):
            if i:
                self.print_string(delimiter)
            print_key(key)
            self.print_string("=")
            print_value(value)

    def _get_new_valid_name_id(self) -> str:
        self._next_valid_name_id[-1] += 1
        return str(self._next_valid_name_id[-1] - 1)

    def _get_new_valid_block_id(self) -> int:
        self._next_valid_block_id[-1] += 1
        return self._next_valid_block_id[-1] - 1

    def _print_results(self, op: Operation) -> None:
        results = op.results
        # No results
        if len(results) == 0:
            return

        # Multiple results
        self.print_list(op.results, self.print)
        self.print_string(" = ")

    def print_ssa_value(self, value: SSAValue) -> str:
        """
        Print an SSA value in the printer. This assigns a name to the value if the value
        does not have one in the current printing context.
        If the value has a name hint, it will use it as a prefix, and otherwise assign
        a number as the name. Numbers are assigned in order.

        Returns the name used for printing the value.
        """
        if value in self._ssa_values:
            name = self._ssa_values[value]
        elif value.name_hint:
            curr_ind = self.ssa_names.get(value.name_hint, 0)
            suffix = f"_{curr_ind}" if curr_ind != 0 else ""
            name = f"{value.name_hint}{suffix}"
            self._ssa_values[value] = name
            self.ssa_names[value.name_hint] = curr_ind + 1
        else:
            name = self._get_new_valid_name_id()
            self._ssa_values[value] = name

        self.print_string(f"%{name}")
        return name

    def print_operand(self, operand: SSAValue) -> None:
        self.print_ssa_value(operand)

    def print_block_name(self, block: Block) -> None:
        self.print_string("^")
        if block not in self.block_names:
            self.block_names[block] = self._get_new_valid_block_id()
        self.print_string(f"{self._block_names[-1][block]}")

    def print_block(
        self,
        block: Block,
        print_block_args: bool = True,
        print_block_terminator: bool = True,
    ) -> None:
        """
        Print a block with syntax `(<caret-ident>`(` <block-args> `)`)? ops* )`
        * If `print_block_args` is False, the label and arguments are not printed.
        * If `print_block_terminator` is False, the block terminator is not printed.
        """

        if print_block_args:
            self._print_new_line()
            self.print_block_name(block)
            if len(block.args) != 0:
                self.print_string("(")
                self.print_list(block.args, self.print_block_argument)
                self.print_string(")")
            self.print_string(":")

        with self.indented():
            for op in block.ops:
                if not print_block_terminator and op.has_trait(
                    IsTerminator, value_if_unregistered=False
                ):
                    continue
                self._print_new_line()
                self.print_op(op)

    def print_block_argument(self, arg: BlockArgument, print_type: bool = True) -> None:
        """
        Print a block argument with its type, e.g. `%arg : i32`
        Optionally, do not print the type.
        """
        self.print_ssa_value(arg)
        if print_type:
            self.print_string(" : ")
            self.print_attribute(arg.type)
            if self.print_debuginfo:
                self.print_string(" loc(unknown)")

    def print_region(
        self,
        region: Region,
        print_entry_block_args: bool = True,
        print_empty_block: bool = True,
        print_block_terminators: bool = True,
    ) -> None:
        """
        Print a region with syntax `{ <block>* }`
        * If `print_entry_block_args` is False, the arguments of the entry block
          are not printed.
        * If `print_empty_block` is False, empty entry blocks are not printed.
        * If `print_block_terminators` is False, the block terminators are not printed.
        """
        # Empty region
        self.print_string("{")
        if (entry_block := region.blocks.first) is None:
            self._print_new_line()
            self.print_string("}")
            return

        print_entry_block_args = (
            bool(entry_block.args) and print_entry_block_args
        ) or (not entry_block.ops and print_empty_block)
        self.print_block(
            entry_block,
            print_block_args=print_entry_block_args,
            print_block_terminator=print_block_terminators,
        )

        next_block = entry_block.next_block
        while next_block is not None:
            self.print_block(next_block, print_block_terminator=print_block_terminators)
            next_block = next_block.next_block

        self._print_new_line()
        self.print_string("}")

    def print_regions(self, regions: Sequence[Region]) -> None:
        if len(regions) == 0:
            return

        self.print_string(" (")
        self.print_list(regions, self.print_region)
        self.print_string(")")

    def print_operands(self, operands: Sequence[SSAValue]) -> None:
        self.print_string("(")
        self.print_list(operands, self.print_operand)
        self.print_string(")")

    def print_paramattr_parameters(
        self, params: Sequence[Attribute], always_print_brackets: bool = False
    ) -> None:
        if len(params) == 0 and not always_print_brackets:
            return
        self.print_string("<")
        self.print_list(params, self.print_attribute)
        self.print_string(">")

    def print_string_literal(self, string: str):
        self.print_string(json.dumps(string))

    def print_identifier_or_string_literal(self, string: str):
        """
        Prints the provided string as an identifier if it is one,
        and as a string literal otherwise.
        """
        if MLIRLexer.bare_identifier_regex.fullmatch(string) is None:
            self.print_string_literal(string)
            return
        self.print_string(string)

    def print_bytes_literal(self, bytestring: bytes):
        self.print_string('"')
        for byte in bytestring:
            match byte:
                case 0x5C:  # ord("\\")
                    self.print_string("\\\\")
                case _ if 0x20 > byte or byte > 0x7E or byte == 0x22:
                    self.print_string(f"\\{byte:02X}")
                case _:
                    self.print_string(chr(byte))
        self.print_string('"')

    def print_float_attr(self, attribute: FloatAttr):
        self.print_float(attribute.value.data, attribute.type)

    def print_float(self, value: float, type: AnyFloat):
        if math.isnan(value) or math.isinf(value):
            if isinstance(type, Float16Type):
                self.print_string(f"{hex(convert_f16_to_u16(value))}")
            elif isinstance(type, Float32Type):
                self.print_string(f"{hex(convert_f32_to_u32(value))}")
            elif isinstance(type, Float64Type):
                self.print_string(f"{hex(convert_f64_to_u64(value))}")
            else:
                raise NotImplementedError(
                    f"Cannot print '{value}' value for float type {str(type)}"
                )
        else:
            # to mirror mlir-opt, attempt to print scientific notation iff the value parses losslessly
            float_str = f"{value:.5e}"
            index = float_str.find("e")
            float_str = float_str[:index] + "0" + float_str[index:]

            parsed_value = type.unpack(type.pack([float(float_str)]), 1)[0]

            if parsed_value == value:
                self.print_string(float_str)
            else:
                if isinstance(type, Float32Type):
                    # f32 is printed with 9 significant digits
                    float_str = f"{value:.9g}"
                    if "." in float_str:
                        self.print_string(float_str)
                    else:
                        self.print_string(f"0x{convert_f32_to_u32(value):X}")
                elif isinstance(type, Float64Type):
                    # f64 is printed with 17 significant digits
                    float_str = f"{value:.17g}"
                    if "." in float_str:
                        self.print_string(float_str)
                    else:
                        self.print_string(f"0x{convert_f64_to_u64(value):X}")
                else:
                    # default to full python precision
                    self.print_string(f"{repr(value)}")

    def print_attribute(self, attribute: Attribute) -> None:
        if isinstance(attribute, UnitAttr):
            self.print_string("unit")
            return

        if isinstance(attribute, LocationAttr):
            self.print_string("loc(unknown)")
            return

        if isinstance(attribute, IntegerType):
            if attribute.signedness.data == Signedness.SIGNLESS:
                self.print_string("i")
            elif attribute.signedness.data == Signedness.SIGNED:
                self.print_string("si")
            elif attribute.signedness.data == Signedness.UNSIGNED:
                self.print_string("ui")
            self.print_string(str(attribute.width.data))
            return

        if isinstance(attribute, BFloat16Type):
            self.print_string("bf16")
            return
        if isinstance(attribute, Float16Type):
            self.print_string("f16")
            return
        if isinstance(attribute, Float32Type):
            self.print_string("f32")
            return
        if isinstance(attribute, Float64Type):
            self.print_string("f64")
            return
        if isinstance(attribute, Float80Type):
            self.print_string("f80")
            return
        if isinstance(attribute, Float128Type):
            self.print_string("f128")
            return

        if isinstance(attribute, IntegerAttr):
            # boolean shorthands
            if (
                isinstance(
                    (ty := attribute.get_type()),
                    IntegerType,
                )
                and ty.width.data == 1
            ):
                self.print_string("true" if attribute.value.data else "false")
                return
            # Otherwise we fall through to TypedAttribute case

        if isinstance(attribute, TypedAttribute):
            attribute.print_without_type(self)
            self.print_string(" : ")
            self.print_attribute(attribute.get_type())
            return

        if isinstance(attribute, StringAttr):
            self.print_string_literal(attribute.data)
            return

        if isinstance(attribute, BytesAttr):
            self.print_bytes_literal(attribute.data)
            return

        if isinstance(attribute, SymbolRefAttr):
            self.print_string("@")
            self.print_identifier_or_string_literal(attribute.root_reference.data)
            for ref in attribute.nested_references.data:
                self.print_string("::@")
                self.print_identifier_or_string_literal(ref.data)
            return

        # Complex types have MLIR shorthands but XDSL does not.
        if isinstance(attribute, ComplexType):
            self.print_string("complex<")
            self.print_attribute(attribute.element_type)
            self.print_string(">")
            return

        if isinstance(attribute, ArrayAttr):
            attribute = cast(ArrayAttr[Attribute], attribute)
            self.print_string("[")
            self.print_list(attribute.data, self.print_attribute)
            self.print_string("]")
            return

        if isinstance(attribute, DenseArrayBase):
            self.print_string("array<")
            self.print_attribute(attribute.elt_type)
            if len(attribute) == 0:
                self.print_string(">")
                return
            data = attribute.iter_values()
            self.print_string(": ")
            # There is a bug in MLIR which will segfault when parsing DenseArrayBase type i1 as 0 or 1,
            # therefore we need to print these as false and true
            if attribute.elt_type == i1:
                self.print_list(
                    data,
                    lambda x: self.print_string("true" if x else "false"),
                )
            else:
                self.print_list(data, lambda x: self.print_string(f"{x}"))
            self.print_string(">")
            return

        if isinstance(attribute, DictionaryAttr):
            self.print_attr_dict(attribute.data)
            return

        if isinstance(attribute, FunctionType):
            self.print_string("(")
            self.print_list(attribute.inputs.data, self.print_attribute)
            self.print_string(") -> ")
            outputs = attribute.outputs.data
            if len(outputs) == 1 and not isinstance(outputs[0], FunctionType):
                self.print_attribute(outputs[0])
            else:
                self.print_string("(")
                self.print_list(outputs, self.print_attribute)
                self.print_string(")")
            return

        if isinstance(attribute, DenseResourceAttr):
            handle = attribute.resource_handle.data
            self.print_string(f"dense_resource<{handle}> : ")
            self.print_attribute(attribute.type)
            return

        if isinstance(attribute, TensorType):
            attribute = cast(AnyVectorType, attribute)
            self.print_string("tensor<")
            self.print_list(
                attribute.shape.data,
                lambda x: (
                    self.print_string(f"{x.data}")
                    if x.data != -1
                    else self.print_string("?")
                ),
                "x",
            )
            if len(attribute.shape.data) != 0:
                self.print_string("x")
            self.print_attribute(attribute.element_type)
            if isinstance(attribute, TensorType) and attribute.encoding != NoneAttr():
                self.print_string(", ")
                self.print_attribute(attribute.encoding)
            self.print_string(">")
            return

        if isinstance(attribute, VectorType):
            attribute = cast(AnyVectorType, attribute)

            self.print_string("vector<")

            self.print_list(
                zip(attribute.shape, attribute.scalable_dims, strict=True),
                self._print_vector_dim,
                delimiter="x",
            )
            if attribute.shape.data:
                self.print_string("x")

            self.print_attribute(attribute.element_type)
            self.print_string(">")
            return

        if isinstance(attribute, UnrankedTensorType):
            attribute = cast(AnyUnrankedTensorType, attribute)
            self.print_string("tensor<*x")
            self.print_attribute(attribute.element_type)
            self.print_string(">")
            return

        if isinstance(attribute, StridedLayoutAttr):
            self.print_string("strided<[")

            def print_int_or_question(value: IntAttr | NoneAttr) -> None:
                self.print_string(
                    f"{value.data}" if isinstance(value, IntAttr) else "?"
                )

            self.print_list(attribute.strides.data, print_int_or_question, ", ")
            self.print_string("]")
            if attribute.offset == IntAttr(0):
                self.print_string(">")
                return
            self.print_string(", offset: ")
            print_int_or_question(attribute.offset)
            self.print_string(">")
            return

        if isinstance(attribute, MemRefType):
            attribute = cast(MemRefType[Attribute], attribute)
            self.print_string("memref<")
            if attribute.shape.data:
                self.print_list(
                    attribute.shape.data,
                    lambda x: (
                        self.print_string(f"{x.data}")
                        if x.data != -1
                        else self.print_string("?")
                    ),
                    "x",
                )
                self.print_string("x")
            self.print_attribute(attribute.element_type)
            if not isinstance(attribute.layout, NoneAttr):
                self.print_string(", ")
                self.print_attribute(attribute.layout)
            if not isinstance(attribute.memory_space, NoneAttr):
                self.print_string(", ")
                self.print_attribute(attribute.memory_space)
            self.print_string(">")
            return

        if isinstance(attribute, UnrankedMemRefType):
            attribute = cast(AnyUnrankedMemRefType, attribute)
            self.print_string("memref<*x")
            self.print_attribute(attribute.element_type)
            if not isinstance(attribute.memory_space, NoneAttr):
                self.print_string(", ")
                self.print_attribute(attribute.memory_space)
            self.print_string(">")
            return

        if isinstance(attribute, IndexType):
            self.print_string("index")
            return

        if isinstance(attribute, NoneType):
            self.print_string("none")
            return

        if isinstance(attribute, OpaqueAttr):
            self.print_string("opaque<")
            self.print_attribute(attribute.ident)
            self.print_string(", ")
            self.print_attribute(attribute.value)
            self.print_string(">")

            if not isinstance(attribute.type, NoneAttr):
                self.print_string(" : ")
                self.print_attribute(attribute.type)
            return

        if isinstance(attribute, AffineMapAttr):
            self.print_string(f"affine_map<{attribute.data}>")
            return

        if isinstance(attribute, AffineSetAttr):
            self.print_string(f"affine_set<{attribute.data}>")
            return

        if isinstance(attribute, UnregisteredAttr):
            # Do not print `!` or `#` for unregistered builtin attributes
            self.print_string("!" if attribute.is_type.data else "#")
            if attribute.is_opaque.data:
                self.print_string(
                    f"{attribute.attr_name.data.replace('.', '<', 1)}{attribute.value.data}>"
                )
            else:
                self.print_string(attribute.attr_name.data)
                if attribute.value.data:
                    self.print_string(f"<{attribute.value.data}>")
            return

        # Print dialect attributes
        self.print_string("!" if isinstance(attribute, TypeAttribute) else "#")

        if isinstance(attribute, OpaqueSyntaxAttribute):
            self.print_string(attribute.name.replace(".", "<", 1))
            if isinstance(attribute, SpacedOpaqueSyntaxAttribute):
                self.print_string(" ")
        else:
            self.print_string(attribute.name)

        if isinstance(attribute, Data):
            attribute.print_parameter(self)

        elif isinstance(attribute, ParametrizedAttribute):
            attribute.print_parameters(self)

        if isinstance(attribute, OpaqueSyntaxAttribute):
            self.print_string(">")

        return

    def _print_vector_dim(self, pair: tuple[IntAttr, BoolAttr]):
        """
        Helper method to print a vector dimension either as static (`4`) or scalable
        (`[4]`).
        """
        dim, scalable = pair
        if scalable:
            self.print_string(f"[{dim.data}]")
        else:
            self.print_string(f"{dim.data}")

    def print_successors(self, successors: Sequence[Block]):
        if len(successors) == 0:
            return
        self.print_string(" [")
        self.print_list(successors, self.print_block_name)
        self.print_string("]")

    def _print_attr_string(self, attr_tuple: tuple[str, Attribute]) -> None:
        # Print the name without quotes if it is a bare identifier
        if MLIRLexer.bare_identifier_regex.fullmatch(attr_tuple[0]):
            self.print_string(attr_tuple[0])
        else:
            self.print_string(f'"{attr_tuple[0]}"')

        if not isinstance(attr_tuple[1], UnitAttr):
            self.print_string(" = ")
            self.print_attribute(attr_tuple[1])

    def print_attr_dict(self, attr_dict: Mapping[str, Attribute]) -> None:
        self.print_string("{")
        self.print_list(attr_dict.items(), self._print_attr_string)
        self.print_string("}")

    def _print_op_properties(self, properties: dict[str, Attribute]) -> None:
        if not properties:
            return

        self.print_string(" ")
        with self.in_angle_brackets():
            self.print_attr_dict(properties)

    def print_op_attributes(
        self,
        attributes: Mapping[str, Attribute],
        *,
        reserved_attr_names: Iterable[str] = (),
        print_keyword: bool = False,
    ) -> bool:
        """
        Prints the attribute dictionary of an operation, with an optional `attributes`
        keyword.
        Values for `reserved_attr_names` are not printed even if present.
        If the printed dictionary would be empty, then nothing is printed, and this
        function returns False.
        """
        if not attributes:
            return False

        if reserved_attr_names:
            attributes = {
                name: attr
                for name, attr in attributes.items()
                if name not in reserved_attr_names
            }

        if not attributes:
            return False

        if print_keyword:
            self.print_string(" attributes")

        self.print_string(" ")
        self.print_attr_dict(attributes)
        return True

    def print_op_with_default_format(self, op: Operation) -> None:
        self.print_operands(op.operands)
        self.print_successors(op.successors)
        if not self.print_properties_as_attributes:
            self._print_op_properties(op.properties)
        self.print_regions(op.regions)
        if self.print_properties_as_attributes:
            clashing_names = op.properties.keys() & op.attributes.keys()
            if clashing_names:
                raise ValueError(
                    f"Properties {', '.join(clashing_names)} would overwrite the attributes of the same names."
                )

            self.print_op_attributes(op.attributes | op.properties)
        else:
            self.print_op_attributes(op.attributes)
        self.print_string(" : ")
        self.print_operation_type(op)

    def print_function_type(
        self, input_types: Iterable[Attribute], output_types: Iterable[Attribute]
    ):
        """
        Prints a function type like `(i32, i64) -> (f32, f64)` with the following
        format:

        The inputs are always a comma-separated list in parentheses.
        If the output has a single element, the parentheses are dropped, except when the
        only return type is a function type, in which case they are kept.

        ```
        () -> ()                 # no inputs, no outputs
        (i32) -> ()              # one input, no outputs
        (i32) -> i32             # one input, one output
        (i32) -> (i32, i32)      # one input, two outputs
        (i32) -> ((i32) -> i32)  # one input, one function type output
        ```
        """
        self.print_string("(")
        self.print_list(input_types, self.print_attribute)
        self.print_string(") -> ")

        remaining_outputs_iterator = iter(output_types)
        try:
            first_type = next(remaining_outputs_iterator)
        except StopIteration:
            # No outputs
            self.print_string("()")
            return

        try:
            second_type = next(remaining_outputs_iterator)
        except StopIteration:
            # One output, drop parentheses unless it's a FunctionType
            if isinstance(first_type, FunctionType):
                self.print_string("(")
                self.print_attribute(first_type)
                self.print_string(")")
            else:
                self.print_attribute(first_type)
            return

        # Two or more outputs, comma-separated list
        self.print_string("(")
        self.print_list(
            chain((first_type, second_type), remaining_outputs_iterator),
            self.print_attribute,
        )
        self.print_string(")")

    def print_operation_type(self, op: Operation) -> None:
        self.print_function_type(op.operand_types, op.result_types)
        if self.print_debuginfo:
            self.print_string(" loc(unknown)")

    def enter_scope(self) -> None:
        self._next_valid_name_id.append(self._next_valid_name_id[-1])
        self._next_valid_block_id.append(self._next_valid_block_id[-1])
        self._ssa_names.append(self._ssa_names[-1].copy())
        self._block_names.append(self._block_names[-1].copy())

    def exit_scope(self) -> None:
        self._next_valid_name_id.pop()
        self._next_valid_block_id.pop()
        self._ssa_names.pop()
        self._block_names.pop()

    def print_op(self, op: Operation) -> None:
        scope = bool(op.get_traits_of_type(IsolatedFromAbove))
        begin_op_pos = self._current_column
        self._print_results(op)
        if scope:
            self.enter_scope()
        use_custom_format = False
        if isinstance(op, UnregisteredOp):
            self.print_string(f'"{op.op_name.data}"')
        # If we print with the generic format, or the operation does not have a custom
        # format
        elif self.print_generic_format or Operation.print is type(op).print:
            self.print_string(f'"{op.name}"')
        else:
            self.print_string(f"{op.name}")
            use_custom_format = True
        end_op_pos = self._current_column
        if op in self.diagnostic.op_messages:
            for message in self.diagnostic.op_messages[op]:
                self._add_message_on_next_line(message, begin_op_pos, end_op_pos)
        if isinstance(op, UnregisteredOp):
            op_name = op.op_name
            del op.attributes["op_name__"]
            self.print_op_with_default_format(op)
            op.attributes["op_name__"] = op_name
        elif use_custom_format:
            op.print(self)
        else:
            self.print_op_with_default_format(op)
        if scope:
            self.exit_scope()



# RUN: python %s | filecheck %s

from xdsl.frontend.pyast.context import CodeContext
from xdsl.frontend.pyast.exception import CodeGenerationException
from xdsl.frontend.pyast.program import FrontendProgram

p = FrontendProgram()
with CodeContext(p):
    # CHECK:      func.func @test_affine_for_I() {
    # CHECK-NEXT:   "affine.for"() <{lowerBoundMap = affine_map<() -> (0)>, upperBoundMap = affine_map<() -> (100)>, step = 1 : index, operandSegmentSizes = array<i32: 0, 0, 0>}> ({
    # CHECK-NEXT:   ^0(%0 : index):
    # CHECK-NEXT:     "affine.yield"() : () -> ()
    # CHECK-NEXT:   }) : () -> ()
    # CHECK-NEXT:   func.return
    # CHECK-NEXT: }

    def test_affine_for_I():
        for _ in range(100):
            pass
        return

    # CHECK:      func.func @test_affine_for_II() {
    # CHECK-NEXT:   "affine.for"() <{lowerBoundMap = affine_map<() -> (10)>, upperBoundMap = affine_map<() -> (30)>, step = 1 : index, operandSegmentSizes = array<i32: 0, 0, 0>}> ({
    # CHECK-NEXT:   ^0(%0 : index):
    # CHECK-NEXT:     "affine.yield"() : () -> ()
    # CHECK-NEXT:   }) : () -> ()
    # CHECK-NEXT:   func.return
    # CHECK-NEXT: }
    def test_affine_for_II():
        for _ in range(10, 30):
            pass
        return

    # CHECK:      func.func @test_affine_for_III() {
    # CHECK-NEXT:   "affine.for"() <{lowerBoundMap = affine_map<() -> (1)>, upperBoundMap = affine_map<() -> (20)>, step = 5 : index, operandSegmentSizes = array<i32: 0, 0, 0>}> ({
    # CHECK-NEXT:   ^0(%0 : index):
    # CHECK-NEXT:     "affine.yield"() : () -> ()
    # CHECK-NEXT:   }) : () -> ()
    # CHECK-NEXT:   func.return
    # CHECK-NEXT: }
    def test_affine_for_III():
        for _ in range(1, 20, 5):
            pass
        return

    # CHECK:      func.func @test_affine_for_IV() {
    # CHECK-NEXT:   "affine.for"() <{lowerBoundMap = affine_map<() -> (0)>, upperBoundMap = affine_map<() -> (10)>, step = 1 : index, operandSegmentSizes = array<i32: 0, 0, 0>}> ({
    # CHECK-NEXT:   ^0(%0 : index):
    # CHECK-NEXT:     "affine.for"() <{lowerBoundMap = affine_map<() -> (0)>, upperBoundMap = affine_map<() -> (20)>, step = 1 : index, operandSegmentSizes = array<i32: 0, 0, 0>}> ({
    # CHECK-NEXT:     ^1(%1 : index):
    # CHECK-NEXT:       "affine.for"() <{lowerBoundMap = affine_map<() -> (0)>, upperBoundMap = affine_map<() -> (30)>, step = 1 : index, operandSegmentSizes = array<i32: 0, 0, 0>}> ({
    # CHECK-NEXT:       ^2(%2 : index):
    # CHECK-NEXT:         "affine.yield"() : () -> ()
    # CHECK-NEXT:       }) : () -> ()
    # CHECK-NEXT:       "affine.yield"() : () -> ()
    # CHECK-NEXT:     }) : () -> ()
    # CHECK-NEXT:     "affine.yield"() : () -> ()
    # CHECK-NEXT:   }) : () -> ()
    # CHECK-NEXT:   func.return
    # CHECK-NEXT: }
    def test_affine_for_IV():
        for _ in range(10):
            for _ in range(20):
                for _ in range(30):
                    pass
        return


p.compile(desymref=False)
print(p.textual_format())


try:
    with CodeContext(p):
        # CHECK: Expected integer constant for loop end, got 'float'.
        def test_not_supported_affine_loop_I():
            for _ in range(
                12.0  # pyright: ignore[reportArgumentType]
            ):
                pass
            return

    p.compile(desymref=False)
    print(p.textual_format())
except CodeGenerationException as e:
    print(e.msg)

try:
    with CodeContext(p):
        # CHECK: Expected integer constant for loop start, got 'str'.
        def test_not_supported_affine_loop_II():
            for _ in range(
                "boom",  # pyright: ignore[reportArgumentType]
                100,
            ):
                pass
            return

    p.compile(desymref=False)
    print(p.textual_format())
except CodeGenerationException as e:
    print(e.msg)

try:
    with CodeContext(p):
        # CHECK: Expected integer constant for loop step, got 'float'.
        def test_not_supported_affine_loop_III():
            for _ in range(
                0,
                100,
                1.0,  # pyright: ignore[reportArgumentType]
            ):
                pass
            return

    p.compile(desymref=False)
    print(p.textual_format())
except CodeGenerationException as e:
    print(e.msg)



from xdsl.context import Context
from xdsl.dialects import builtin
from xdsl.dialects.stencil import ApplyOp, BufferOp, CombineOp, StoreOp
from xdsl.ir import OpResult, SSAValue
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)


def should_materialize(temp: SSAValue):
    """
    Predicates if a specific stencil.apply output should be buffered.
    It should if it is used by another stencil.apply and not already buffered or stored.
    """
    return any(isinstance(u.operation, ApplyOp) for u in temp.uses) and not any(
        isinstance(u.operation, StoreOp | BufferOp) for u in temp.uses
    )


class ApplyOpMaterialization(RewritePattern):
    """
    Adds stencil.buffer to any used output of a stencil.apply that is not otherwised
    mapped to storage.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ApplyOp | CombineOp, rewriter: PatternRewriter, /):
        clone = op.clone()
        new_res: list[OpResult] = []
        buffers: list[BufferOp] = []
        for i, out in enumerate(op.results):
            if should_materialize(out):
                buffer = BufferOp(clone.results[i])
                buffers.append(buffer)
                new_res.append(buffer.res)
            else:
                new_res.append(out)
        if buffers:
            rewriter.replace_matched_op([clone, *buffers], new_res)
        else:
            clone.erase()


class StencilStorageMaterializationPass(ModulePass):
    """
    Pass adding stencil.buffer whenever necessary to lower a stencil dialect IR,
    by adding stencil.buffer on any used stencil.apply output not otherwise mapped
    to storage.
    """

    name = "stencil-storage-materialization"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    ApplyOpMaterialization(),
                ]
            )
        ).rewrite_module(op)



from xdsl.builder import ImplicitBuilder
from xdsl.dialects import test
from xdsl.dialects.builtin import IndexType
from xdsl.ir import Block, Region
from xdsl.transforms.loop_invariant_code_motion import can_be_hoisted

index = IndexType()


def test_can_be_hoisted():
    """Test for with loop-variant variables"""

    outer_region = Region(Block(arg_types=(index,)))
    mid_region = Region(Block(arg_types=(index,)))
    inner_region = Region(Block(arg_types=(index,)))

    with ImplicitBuilder(outer_region) as (outer_arg,):
        outer_op_no_args = test.TestOp()
        outer_op_outer_arg = test.TestOp(operands=(outer_arg,))
        with ImplicitBuilder(mid_region) as (mid_arg,):
            mid_op_no_args = test.TestOp()
            mid_op_outer_arg = test.TestOp(operands=(outer_arg,))
            mid_op_mid_arg = test.TestOp(operands=(mid_arg,))
            with ImplicitBuilder(inner_region) as (inner_arg,):
                inner_op_no_args = test.TestOp()
                inner_op_outer_arg = test.TestOp(operands=(outer_arg,))
                inner_op_inner_arg = test.TestOp(operands=(inner_arg,))
                inner_term_op = test.TestTermOp()
            mid_term_op = test.TestTermOp()
            inner_region_op = test.TestOp(regions=(inner_region,))
        mid_region_op = test.TestOp(regions=(mid_region,))
        out_term_op = test.TestTermOp()

    # outer region
    assert can_be_hoisted(outer_op_no_args, outer_region)
    assert not can_be_hoisted(outer_op_outer_arg, outer_region)
    assert can_be_hoisted(mid_op_no_args, outer_region)
    assert not can_be_hoisted(mid_op_outer_arg, outer_region)
    assert not can_be_hoisted(mid_op_mid_arg, outer_region)
    assert can_be_hoisted(inner_op_no_args, outer_region)
    assert not can_be_hoisted(inner_op_outer_arg, outer_region)
    assert not can_be_hoisted(inner_op_inner_arg, outer_region)
    assert not can_be_hoisted(inner_term_op, outer_region)
    assert not can_be_hoisted(mid_term_op, outer_region)
    assert not can_be_hoisted(inner_region_op, outer_region)
    assert not can_be_hoisted(mid_region_op, outer_region)
    assert not can_be_hoisted(out_term_op, outer_region)

    # mid region
    assert can_be_hoisted(outer_op_no_args, mid_region)
    assert can_be_hoisted(outer_op_outer_arg, mid_region)
    assert can_be_hoisted(mid_op_no_args, mid_region)
    assert can_be_hoisted(mid_op_outer_arg, mid_region)
    assert not can_be_hoisted(mid_op_mid_arg, mid_region)
    assert can_be_hoisted(inner_op_no_args, mid_region)
    assert can_be_hoisted(inner_op_outer_arg, mid_region)
    assert not can_be_hoisted(inner_op_inner_arg, mid_region)
    assert not can_be_hoisted(inner_term_op, mid_region)
    assert not can_be_hoisted(mid_term_op, mid_region)
    assert can_be_hoisted(inner_region_op, mid_region)
    assert can_be_hoisted(mid_region_op, mid_region)
    assert not can_be_hoisted(out_term_op, mid_region)

    # inner region
    assert can_be_hoisted(outer_op_no_args, inner_region)
    assert can_be_hoisted(outer_op_outer_arg, inner_region)
    assert can_be_hoisted(mid_op_no_args, inner_region)
    assert can_be_hoisted(mid_op_outer_arg, inner_region)
    assert can_be_hoisted(mid_op_mid_arg, inner_region)
    assert can_be_hoisted(inner_op_no_args, inner_region)
    assert can_be_hoisted(inner_op_outer_arg, inner_region)
    assert not can_be_hoisted(inner_op_inner_arg, inner_region)
    assert not can_be_hoisted(inner_term_op, inner_region)
    assert not can_be_hoisted(mid_term_op, inner_region)
    assert can_be_hoisted(inner_region_op, inner_region)
    assert can_be_hoisted(mid_region_op, inner_region)
    assert not can_be_hoisted(out_term_op, inner_region)



from __future__ import annotations

from collections.abc import Sequence
from typing import ClassVar

from xdsl.dialects.builtin import (
    IndexType,
    IndexTypeConstr,
    MemRefType,
    SignlessIntegerConstraint,
    VectorBaseTypeAndRankConstraint,
    VectorBaseTypeConstraint,
    VectorRankConstraint,
    VectorType,
    i1,
)
from xdsl.ir import Attribute, Dialect, Operation, SSAValue
from xdsl.irdl import (
    IRDLOperation,
    VarConstraint,
    base,
    irdl_op_definition,
    operand_def,
    opt_operand_def,
    result_def,
    traits_def,
    var_operand_def,
)
from xdsl.traits import Pure
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.hints import assert_isa, isa


@irdl_op_definition
class LoadOp(IRDLOperation):
    name = "vector.load"
    base = operand_def(MemRefType)
    indices = var_operand_def(IndexType)
    result = result_def(VectorType)

    assembly_format = (
        "$base `[` $indices `]` attr-dict `:` type($base) `,` type($result)"
    )

    def verify_(self):
        assert isa(self.base.type, MemRefType[Attribute])
        assert isa(self.result.type, VectorType[Attribute])

        if self.base.type.element_type != self.result.type.element_type:
            raise VerifyException(
                "MemRef element type should match the Vector element type."
            )

        if self.base.type.get_num_dims() != len(self.indices):
            raise VerifyException("Expected an index for each dimension.")

    @staticmethod
    def get(
        ref: SSAValue | Operation, indices: Sequence[SSAValue | Operation]
    ) -> LoadOp:
        ref = SSAValue.get(ref)
        assert assert_isa(ref.type, MemRefType[Attribute])

        return LoadOp.build(
            operands=[ref, indices],
            result_types=[VectorType(ref.type.element_type, [1])],
        )


@irdl_op_definition
class StoreOp(IRDLOperation):
    name = "vector.store"
    vector = operand_def(VectorType)
    base = operand_def(MemRefType)
    indices = var_operand_def(IndexType)

    assembly_format = (
        "$vector `,` $base `[` $indices `]` attr-dict `:` type($base) `,` type($vector)"
    )

    def verify_(self):
        assert isa(self.base.type, MemRefType[Attribute])
        assert isa(self.vector.type, VectorType[Attribute])

        if self.base.type.element_type != self.vector.type.element_type:
            raise VerifyException(
                "MemRef element type should match the Vector element type."
            )

        if self.base.type.get_num_dims() != len(self.indices):
            raise VerifyException("Expected an index for each dimension.")

    @staticmethod
    def get(
        vector: Operation | SSAValue,
        ref: Operation | SSAValue,
        indices: Sequence[Operation | SSAValue],
    ) -> StoreOp:
        return StoreOp.build(operands=[vector, ref, indices])


@irdl_op_definition
class BroadcastOp(IRDLOperation):
    name = "vector.broadcast"
    source = operand_def()
    vector = result_def(VectorType)
    traits = traits_def(Pure())

    assembly_format = "$source attr-dict `:` type($source) `to` type($vector)"

    def verify_(self):
        assert isa(self.vector.type, VectorType[Attribute])

        if self.source.type != self.vector.type.element_type:
            raise VerifyException(
                "Source operand and result vector must have the same element type."
            )

    @staticmethod
    def get(source: Operation | SSAValue) -> BroadcastOp:
        return BroadcastOp.build(
            operands=[source],
            result_types=[VectorType(SSAValue.get(source).type, [1])],
        )


@irdl_op_definition
class FMAOp(IRDLOperation):
    name = "vector.fma"

    T: ClassVar = VarConstraint("T", base(VectorType))

    lhs = operand_def(T)
    rhs = operand_def(T)
    acc = operand_def(T)
    res = result_def(T)
    traits = traits_def(Pure())

    assembly_format = "$lhs `,` $rhs `,` $acc attr-dict `:` type($lhs)"

    @staticmethod
    def get(
        lhs: Operation | SSAValue, rhs: Operation | SSAValue, acc: Operation | SSAValue
    ) -> FMAOp:
        lhs = SSAValue.get(lhs)
        assert assert_isa(lhs.type, VectorType[Attribute])

        return FMAOp.build(
            operands=[lhs, rhs, acc],
            result_types=[VectorType(lhs.type.element_type, [1])],
        )


@irdl_op_definition
class MaskedLoadOp(IRDLOperation):
    name = "vector.maskedload"
    base = operand_def(MemRefType)
    indices = var_operand_def(IndexType)
    mask = operand_def(VectorBaseTypeAndRankConstraint(i1, 1))
    pass_thru = operand_def(VectorType)
    result = result_def(VectorRankConstraint(1))

    assembly_format = "$base `[` $indices `]` `,` $mask `,` $pass_thru attr-dict `:` type($base) `,` type($mask) `,` type($pass_thru) `into` type($result)"  # noqa: E501

    def verify_(self):
        memref_type = self.base.type
        assert isa(memref_type, MemRefType[Attribute])
        memref_element_type = memref_type.element_type

        res_type = self.result.type
        assert isa(res_type, VectorType[Attribute])
        res_element_type = res_type.element_type

        passthrough_type = self.pass_thru.type
        assert isa(passthrough_type, VectorType[Attribute])
        passthrough_element_type = passthrough_type.element_type

        if memref_element_type != res_element_type:
            raise VerifyException(
                "MemRef element type should match the result vector and passthrough vector "
                "element type. Found different element types for memref and result."
            )
        elif memref_element_type != passthrough_element_type:
            raise VerifyException(
                "MemRef element type should match the result vector and passthrough vector "
                "element type. Found different element types for memref and passthrough."
            )

        if memref_type.get_num_dims() != len(self.indices):
            raise VerifyException("Expected an index for each memref dimension.")

    @staticmethod
    def get(
        memref: SSAValue | Operation,
        indices: Sequence[SSAValue | Operation],
        mask: SSAValue | Operation,
        passthrough: SSAValue | Operation,
    ) -> MaskedLoadOp:
        memref = SSAValue.get(memref)
        assert assert_isa(memref.type, MemRefType[Attribute])

        return MaskedLoadOp.build(
            operands=[memref, indices, mask, passthrough],
            result_types=[VectorType(memref.type.element_type, [1])],
        )


@irdl_op_definition
class MaskedStoreOp(IRDLOperation):
    name = "vector.maskedstore"
    base = operand_def(MemRefType)
    indices = var_operand_def(IndexType)
    mask = operand_def(VectorBaseTypeAndRankConstraint(i1, 1))
    value_to_store = operand_def(VectorRankConstraint(1))

    assembly_format = "$base `[` $indices `]` `,` $mask `,` $value_to_store attr-dict `:` type($base) `,` type($mask) `,` type($value_to_store)"  # noqa: E501

    def verify_(self):
        memref_type = self.base.type
        assert isa(memref_type, MemRefType[Attribute])
        memref_element_type = memref_type.element_type

        value_to_store_type = self.value_to_store.type
        assert isa(value_to_store_type, VectorType[Attribute])

        mask_type = self.mask.type
        assert isa(mask_type, VectorType[Attribute])

        if memref_element_type != value_to_store_type.element_type:
            raise VerifyException(
                "MemRef element type should match the stored vector type. "
                "Obtained types were "
                + str(memref_element_type)
                + " and "
                + str(value_to_store_type.element_type)
                + "."
            )

        if memref_type.get_num_dims() != len(self.indices):
            raise VerifyException("Expected an index for each memref dimension.")

    @staticmethod
    def get(
        memref: SSAValue | Operation,
        indices: Sequence[SSAValue | Operation],
        mask: SSAValue | Operation,
        value_to_store: SSAValue | Operation,
    ) -> MaskedStoreOp:
        return MaskedStoreOp.build(operands=[memref, indices, mask, value_to_store])


@irdl_op_definition
class PrintOp(IRDLOperation):
    name = "vector.print"
    source = operand_def()

    @staticmethod
    def get(source: Operation | SSAValue) -> PrintOp:
        return PrintOp.build(operands=[source])


@irdl_op_definition
class CreateMaskOp(IRDLOperation):
    name = "vector.create_mask"
    mask_dim_sizes = var_operand_def(IndexType)
    mask_vector = result_def(VectorBaseTypeConstraint(i1))

    assembly_format = "$mask_dim_sizes attr-dict `:` type(results)"

    def verify_(self):
        assert isa(self.mask_vector.type, VectorType[Attribute])
        if self.mask_vector.type.get_num_dims() != len(self.mask_dim_sizes):
            raise VerifyException(
                "Expected an operand value for each dimension of resultant mask."
            )

    @staticmethod
    def get(mask_operands: list[Operation | SSAValue]) -> CreateMaskOp:
        return CreateMaskOp.build(
            operands=[mask_operands],
            result_types=[VectorType(i1, [1])],
        )


@irdl_op_definition
class ExtractElementOp(IRDLOperation):
    name = "vector.extractelement"
    vector = operand_def(VectorType)
    position = opt_operand_def(IndexTypeConstr | SignlessIntegerConstraint)
    result = result_def(Attribute)
    traits = traits_def(Pure())

    def verify_(self):
        assert isa(self.vector.type, VectorType[Attribute])

        if self.result.type != self.vector.type.element_type:
            raise VerifyException(
                "Expected result type to match element type of vector operand."
            )

        if self.vector.type.get_num_dims() == 0:
            if self.position is not None:
                raise VerifyException("Expected position to be empty with 0-D vector.")
            return
        if self.vector.type.get_num_dims() != 1:
            raise VerifyException("Unexpected >1 vector rank.")
        if self.position is None:
            raise VerifyException("Expected position for 1-D vector.")

    def __init__(
        self,
        vector: SSAValue | Operation,
        position: SSAValue | Operation | None = None,
    ):
        vector = SSAValue.get(vector)
        assert isa(vector.type, VectorType[Attribute])

        result_type = vector.type.element_type

        super().__init__(
            operands=[vector, position],
            result_types=[result_type],
        )


@irdl_op_definition
class InsertElementOp(IRDLOperation):
    name = "vector.insertelement"
    source = operand_def(Attribute)
    dest = operand_def(VectorType)
    position = opt_operand_def(IndexTypeConstr | SignlessIntegerConstraint)
    result = result_def(VectorType)
    traits = traits_def(Pure())

    def verify_(self):
        assert isa(self.dest.type, VectorType[Attribute])

        if self.result.type != self.dest.type:
            raise VerifyException(
                "Expected dest operand and result to have matching types."
            )
        if self.source.type != self.dest.type.element_type:
            raise VerifyException(
                "Expected source operand type to match element type of dest operand."
            )

        if self.dest.type.get_num_dims() == 0:
            if self.position is not None:
                raise VerifyException("Expected position to be empty with 0-D vector.")
            return
        if self.dest.type.get_num_dims() != 1:
            raise VerifyException("Unexpected >1 vector rank.")
        if self.position is None:
            raise VerifyException("Expected position for 1-D vector.")

    def __init__(
        self,
        source: SSAValue | Operation,
        dest: SSAValue | Operation,
        position: SSAValue | Operation | None = None,
    ):
        dest = SSAValue.get(dest)
        assert isa(dest.type, VectorType[Attribute])

        result_type = SSAValue.get(dest).type

        super().__init__(
            operands=[source, dest, position],
            result_types=[result_type],
        )


Vector = Dialect(
    "vector",
    [
        LoadOp,
        StoreOp,
        BroadcastOp,
        FMAOp,
        MaskedLoadOp,
        MaskedStoreOp,
        PrintOp,
        CreateMaskOp,
        ExtractElementOp,
        InsertElementOp,
    ],
    [],
)



from textual import on
from textual.app import ComposeResult
from textual.containers import ScrollableContainer
from textual.screen import Screen
from textual.widgets import (
    Button,
    DirectoryTree,
)


class LoadFile(Screen[str]):
    CSS_PATH = "load_file_screen.tcss"

    directory_tree: DirectoryTree

    def __init__(self):
        self.directory_tree = DirectoryTree("./", id="directory_tree")
        super().__init__()

    def compose(self) -> ComposeResult:
        with ScrollableContainer(id="directory_tree_container"):
            yield self.directory_tree
            yield Button("Cancel", id="quit_load_file_screen_button")

    def on_mount(self) -> None:
        """Configure widgets in this application before it is first shown."""
        self.query_one("#directory_tree_container").border_title = "Click File to Open"

    def on_directory_tree_file_selected(
        self, event: DirectoryTree.FileSelected
    ) -> None:
        selected_path = str(event.path)
        self.dismiss(selected_path)

    @on(Button.Pressed, "#quit_load_file_screen_button")
    def exit_screen(self, event: Button.Pressed) -> None:
        self.dismiss()



from __future__ import annotations

from collections.abc import Callable
from typing import Any, Generic, Literal, TypeAlias, TypeVar

import xdsl.dialects.builtin as builtin
from xdsl.dialects.builtin import Signedness


class _FrontendType:
    """Represents any type in the frontend."""

    @staticmethod
    def to_xdsl() -> Callable[[], Any]:
        raise NotImplementedError()


# Type parameters for integers.
_Width = TypeVar("_Width", bound=int)
_Signedness = TypeVar("_Signedness", bound=Signedness)


# Note the types ignored below:
# a) on each function, since the functions are constrained on a limited set of
#    known types, and _Integer can represent types outside of that set.
# b) on functions that return `bool` in object, instead of `i1`
class _Integer(Generic[_Width, _Signedness], _FrontendType):
    """
    Represents an integer type in the frontend. Should not be used explicitly.
    """

    @staticmethod
    def to_xdsl() -> Callable[[], Any]:
        return builtin.IntegerType.__call__

    def __add__(
        self, other: _Integer[_Width, _Signedness]
    ) -> _Integer[_Width, _Signedness]:
        from xdsl.frontend.pyast.dialects.arith import addi

        return addi(
            self,  # pyright: ignore[reportUnknownVariableType, reportArgumentType]
            other,  # pyright: ignore[reportArgumentType]
        )

    def __and__(
        self, other: _Integer[_Width, _Signedness]
    ) -> _Integer[_Width, _Signedness]:
        from xdsl.frontend.pyast.dialects.arith import andi

        return andi(
            self,  # pyright: ignore[reportUnknownVariableType, reportArgumentType]
            other,  # pyright: ignore[reportArgumentType]
        )

    def __lshift__(
        self, other: _Integer[_Width, _Signedness]
    ) -> _Integer[_Width, _Signedness]:
        from xdsl.frontend.pyast.dialects.arith import shli

        return shli(
            self,  # pyright: ignore[reportUnknownVariableType, reportArgumentType]
            other,  # pyright: ignore[reportArgumentType]
        )

    def __mul__(
        self, other: _Integer[_Width, _Signedness]
    ) -> _Integer[_Width, _Signedness]:
        from xdsl.frontend.pyast.dialects.arith import muli

        return muli(
            self,  # pyright: ignore[reportUnknownVariableType, reportArgumentType]
            other,  # pyright: ignore[reportArgumentType]
        )

    def __rshift__(
        self, other: _Integer[_Width, _Signedness]
    ) -> _Integer[_Width, _Signedness]:
        from xdsl.frontend.pyast.dialects.arith import shrsi

        return shrsi(
            self,  # pyright: ignore[reportUnknownVariableType, reportArgumentType]
            other,  # pyright: ignore[reportArgumentType]
        )

    def __sub__(
        self, other: _Integer[_Width, _Signedness]
    ) -> _Integer[_Width, _Signedness]:
        from xdsl.frontend.pyast.dialects.arith import subi

        return subi(
            self,  # pyright: ignore[reportUnknownVariableType, reportArgumentType]
            other,  # pyright: ignore[reportArgumentType]
        )

    def __eq__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, other: _Integer[_Width, _Signedness]
    ) -> i1:
        from xdsl.frontend.pyast.dialects.arith import cmpi

        return cmpi(
            self,  # pyright: ignore[reportArgumentType]
            other,  # pyright: ignore[reportArgumentType]
            "eq",
        )

    def __ge__(self, other: _Integer[_Width, _Signedness]) -> i1:
        from xdsl.frontend.pyast.dialects.arith import cmpi

        return cmpi(
            self,  # pyright: ignore[reportArgumentType]
            other,  # pyright: ignore[reportArgumentType]
            "sge",
        )

    def __gt__(self, other: _Integer[_Width, _Signedness]) -> i1:
        from xdsl.frontend.pyast.dialects.arith import cmpi

        return cmpi(
            self,  # pyright: ignore[reportArgumentType]
            other,  # pyright: ignore[reportArgumentType]
            "sgt",
        )

    def __le__(self, other: _Integer[_Width, _Signedness]) -> i1:
        from xdsl.frontend.pyast.dialects.arith import cmpi

        return cmpi(
            self,  # pyright: ignore[reportArgumentType]
            other,  # pyright: ignore[reportArgumentType]
            "sle",
        )

    def __lt__(self, other: _Integer[_Width, _Signedness]) -> i1:
        from xdsl.frontend.pyast.dialects.arith import cmpi

        return cmpi(
            self,  # pyright: ignore[reportArgumentType]
            other,  # pyright: ignore[reportArgumentType]
            "slt",
        )

    def __ne__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, other: _Integer[_Width, _Signedness]
    ) -> i1:
        from xdsl.frontend.pyast.dialects.arith import cmpi

        return cmpi(
            self,  # pyright: ignore[reportArgumentType]
            other,  # pyright: ignore[reportArgumentType]
            "ne",
        )


# Type aliases for signless integers.
i1: TypeAlias = _Integer[Literal[1], Literal[Signedness.SIGNLESS]]
i32: TypeAlias = _Integer[Literal[32], Literal[Signedness.SIGNLESS]]
i64: TypeAlias = _Integer[Literal[64], Literal[Signedness.SIGNLESS]]

# Type aliases for signed integers.
si32: TypeAlias = _Integer[Literal[32], Literal[Signedness.SIGNED]]
si64: TypeAlias = _Integer[Literal[64], Literal[Signedness.SIGNED]]

# Type aliases for unsigned integers.
ui32: TypeAlias = _Integer[Literal[32], Literal[Signedness.UNSIGNED]]
ui64: TypeAlias = _Integer[Literal[64], Literal[Signedness.UNSIGNED]]


class _Index(_FrontendType):
    """
    Represents an index type in the frontend. Should not be used explicitly.
    """

    @staticmethod
    def to_xdsl() -> Callable[[], Any]:
        return builtin.IndexType.__call__


# Type alias for index type.
index: TypeAlias = _Index


class _Float16(_FrontendType):
    """
    Represents a 16-bit floating-point type in the frontend. Should not be used
    explicitly.
    """

    @staticmethod
    def to_xdsl() -> Callable[[], Any]:
        return builtin.Float16Type.__call__

    def __add__(self, other: f16) -> f16:
        from xdsl.frontend.pyast.dialects.arith import addf

        return addf(self, other)

    def __sub__(self, other: f16) -> f16:
        from xdsl.frontend.pyast.dialects.arith import subf

        return subf(self, other)

    def __mul__(self, other: f16) -> f16:
        from xdsl.frontend.pyast.dialects.arith import mulf

        return mulf(self, other)


class _Float32(_FrontendType):
    """
    Represents a 32-bit floating-point type in the frontend. Should not be used
    explicitly.
    """

    @staticmethod
    def to_xdsl() -> Callable[[], Any]:
        return builtin.Float32Type.__call__

    def __add__(self, other: f32) -> f32:
        from xdsl.frontend.pyast.dialects.arith import addf

        return addf(self, other)

    def __sub__(self, other: f32) -> f32:
        from xdsl.frontend.pyast.dialects.arith import subf

        return subf(self, other)

    def __mul__(self, other: f32) -> f32:
        from xdsl.frontend.pyast.dialects.arith import mulf

        return mulf(self, other)


class _Float64(_FrontendType):
    """
    Represents a 64-bit floating-point type in the frontend. Should not be used
    explicitly.
    """

    @staticmethod
    def to_xdsl() -> Callable[[], Any]:
        return builtin.Float64Type.__call__

    def __add__(self, other: f64) -> f64:
        from xdsl.frontend.pyast.dialects.arith import addf

        return addf(self, other)

    def __sub__(self, other: f64) -> f64:
        from xdsl.frontend.pyast.dialects.arith import subf

        return subf(self, other)

    def __mul__(self, other: f64) -> f64:
        from xdsl.frontend.pyast.dialects.arith import mulf

        return mulf(self, other)


# Type alias for floating-point types.
f16: TypeAlias = _Float16
f32: TypeAlias = _Float32
f64: TypeAlias = _Float64



from collections import Counter
from collections.abc import Iterable, Iterator, Sequence

from xdsl.dialects import builtin, riscv
from xdsl.ir import Attribute, Block, Operation, SSAValue
from xdsl.pattern_rewriter import PatternRewriter
from xdsl.rewriter import InsertPoint


def register_type_for_type(
    attr: Attribute,
) -> type[riscv.IntRegisterType] | type[riscv.FloatRegisterType]:
    """
    Returns the appropriate register fype for a given input type.
    """
    if isinstance(attr, riscv.IntRegisterType | riscv.FloatRegisterType):
        return type(attr)
    if isinstance(attr, builtin.AnyFloat):
        return riscv.FloatRegisterType
    return riscv.IntRegisterType


def cast_to_regs(values: Iterable[SSAValue]) -> tuple[list[Operation], list[SSAValue]]:
    """
    Return cast operations for operands that don't already have a register type, and
    the new list of values that are all guaranteed to have register types.
    """

    new_ops = list[Operation]()
    new_values = list[SSAValue]()

    for value in values:
        if not isinstance(value.type, riscv.RISCVRegisterType):
            register_type = register_type_for_type(value.type)
            cast_op = builtin.UnrealizedConversionCastOp.get(
                (value,), (register_type.unallocated(),)
            )
            new_ops.append(cast_op)
            value = cast_op.results[0]
        new_values.append(value)

    return new_ops, new_values


def move_ops_for_value(
    value: SSAValue, value_type: Attribute, rd: riscv.RISCVRegisterType
) -> tuple[Operation, SSAValue]:
    """
    Returns the operation that moves the value from the input to a new register.
    In order to disambiguate which floating point move should be used (fmv.s vs fmv.d),
    the floating point type in question must be passed
    """

    if isinstance(rd, riscv.IntRegisterType):
        mv_op = riscv.MVOp(value, rd=rd)
        return mv_op, mv_op.rd
    elif isinstance(rd, riscv.FloatRegisterType):
        match value_type:
            case builtin.Float64Type():
                mv_op = riscv.FMvDOp(value, rd=rd)
            case builtin.Float32Type():
                mv_op = riscv.FMVOp(value, rd=rd)
            case _:
                raise NotImplementedError(
                    f"Move operation for float register containing value of type {value.type} is not implemented"
                )
        return mv_op, mv_op.rd
    else:
        raise NotImplementedError(f"Unsupported register type for move op: {rd}")


def move_to_regs(
    values: Iterable[SSAValue],
    value_types: Iterable[Attribute],
    reg_types: Iterable[riscv.RISCVRegisterType],
) -> tuple[list[Operation], list[SSAValue]]:
    """
    Return move operations to `a` registers (a0, a1, ... | fa0, fa1, ...).
    """

    new_ops = list[Operation]()
    new_values = list[SSAValue]()

    for value, value_type, register_type in zip(
        values, value_types, reg_types, strict=True
    ):
        move_op, new_value = move_ops_for_value(value, value_type, register_type)
        new_ops.append(move_op)
        new_values.append(new_value)

    return new_ops, new_values


def a_regs_for_types(types: Iterable[Attribute]) -> Iterator[riscv.RISCVRegisterType]:
    """
    Returns the "a" registers in which to store types, i.e. `fa0`, `fa1`, etc for
    floating-point values and `a0`, `a1`, etc for integer values and pointers. The
    register index is separate for integer and floating-point registers according to the
    RISC-V ABI.
    """
    counter = Counter[type[riscv.RISCVRegisterType]]()
    for attr_type in types:
        register_type = register_type_for_type(attr_type)
        index = counter[register_type]
        yield register_type.a_register(index)
        counter[register_type] += 1


def a_regs(values: Iterable[SSAValue]) -> Iterator[riscv.RISCVRegisterType]:
    return a_regs_for_types(value.type for value in values)


def move_to_a_regs(
    values: Iterable[SSAValue],
    value_types: Iterable[Attribute],
) -> tuple[list[Operation], list[SSAValue]]:
    """
    Return move operations to `a` registers (a0, a1, ... | fa0, fa1, ...).
    """
    return move_to_regs(values, value_types, a_regs(values))


def move_to_unallocated_regs(
    values: Iterable[SSAValue],
    value_types: Iterable[Attribute],
) -> tuple[list[Operation], list[SSAValue]]:
    """
    Return move operations to unallocated registers.
    """

    new_ops = list[Operation]()
    new_values = list[SSAValue]()

    for value, value_type in zip(values, value_types, strict=True):
        register_type = register_type_for_type(value.type)
        move_op, new_value = move_ops_for_value(
            value, value_type, register_type.unallocated()
        )
        new_ops.append(move_op)
        new_values.append(new_value)

    return new_ops, new_values


def cast_ops_for_values(
    values: Sequence[SSAValue],
) -> tuple[list[Operation], list[SSAValue]]:
    """
    Returns cast operations and new SSA values. The SSA values are guaranteed to be either
    the original SSA value, if it already had a register type, or the result of a cast
    operation. The resulting list has the same length and same order as the input.
    """

    new_ops = list[Operation]()
    new_values = list[SSAValue]()

    for value in values:
        if not isinstance(value.type, riscv.IntRegisterType | riscv.FloatRegisterType):
            new_type = register_type_for_type(value.type)
            cast_op = builtin.UnrealizedConversionCastOp.get(
                (value,), (new_type.unallocated(),)
            )
            new_ops.append(cast_op)
            new_value = cast_op.results[0]
            new_value.name_hint = value.name_hint
        else:
            new_value = value

        new_values.append(new_value)

    return new_ops, new_values


def cast_operands_to_regs(rewriter: PatternRewriter) -> list[SSAValue]:
    """
    Add cast operations just before the targeted operation
    if the operands were not already int registers
    """

    new_ops, new_operands = cast_ops_for_values(rewriter.current_operation.operands)
    rewriter.insert_op_before_matched_op(new_ops)
    return new_operands


def cast_matched_op_results(rewriter: PatternRewriter) -> list[SSAValue]:
    """
    Add cast operations just after the matched operation, to preserve the type validity of
    arguments of uses of results.
    """

    results = [
        builtin.UnrealizedConversionCastOp.get((val,), (val.type,))
        for val in rewriter.current_operation.results
    ]

    for res, result in zip(rewriter.current_operation.results, results):
        for use in set(res.uses):
            # avoid recursion on the casts we just inserted
            if use.operation != result:
                use.operation.operands[use.index] = result.results[0]

    rewriter.insert_op_after_matched_op(results)
    return [result.results[0] for result in results]


def cast_block_args_from_a_regs(block: Block, rewriter: PatternRewriter):
    """
    Change the type of the block arguments to "a" registers and add cast operations just
    after the block entry. Use `fa0`, `fa1`, etc for floating-point values and `a0`, `a1`,
    etc for integer values and pointers. The register index is separate for integer and
    floating-point registers according to the RISC-V ABI.
    """

    new_ops: list[Operation] = []
    counter = Counter[type[riscv.RISCVRegisterType]]()

    for arg in block.args:
        register_type = register_type_for_type(arg.type)
        move_op, new_value = move_ops_for_value(
            arg, arg.type, register_type.unallocated()
        )
        cast_op = builtin.UnrealizedConversionCastOp.get((new_value,), (arg.type,))
        new_ops.append(move_op)
        new_ops.append(cast_op)

        index = counter[register_type]
        arg.type = register_type.a_register(index)
        counter[register_type] += 1
        arg.replace_by(cast_op.results[0])
        move_op.operands[0] = arg

    rewriter.insert_op(new_ops, InsertPoint.at_start(block))


def cast_block_args_to_regs(block: Block, rewriter: PatternRewriter):
    """
    Change the type of the block arguments to registers and add cast operations just after
    the block entry.
    """

    for arg in block.args:
        rewriter.insert_op(
            new_val := builtin.UnrealizedConversionCastOp(
                operands=[arg], result_types=[arg.type]
            ),
            InsertPoint.at_start(block),
        )

        arg.type = register_type_for_type(arg.type).unallocated()
        arg.replace_by(new_val.results[0])
        new_val.operands[new_val.results[0].index] = arg



from xdsl.builder import ImplicitBuilder
from xdsl.context import Context
from xdsl.dialects import arith, func, get_all_dialects
from xdsl.dialects.builtin import (
    IndexType,
    IntegerAttr,
    ModuleOp,
)
from xdsl.interactive.pass_metrics import (
    count_number_of_operations,
    get_diff_operation_count,
)
from xdsl.ir import Block, Region
from xdsl.parser import Parser


def test_operation_counter():
    # build module
    index = IndexType()
    module = ModuleOp(Region([Block()]))
    with ImplicitBuilder(module.body):
        function = func.FuncOp("hello", ((index,), (index,)))
        with ImplicitBuilder(function.body) as (n,):
            two = arith.ConstantOp(IntegerAttr(2, index)).result
            three = arith.ConstantOp(IntegerAttr(2, index)).result
            res_1 = arith.MuliOp(n, two)
            res_2 = arith.MuliOp(n, three)
            res = arith.MuliOp(res_1, res_2)
            func.ReturnOp(res)

    expected_res = {
        "func.func": 1,
        "arith.constant": 2,
        "arith.muli": 3,
        "func.return": 1,
        "builtin.module": 1,
    }

    res = count_number_of_operations(module)
    assert res == expected_res


def test_operation_counter_with_parsing_text():
    text = """builtin.module {
  func.func @hello(%n : index) -> index {
    %two = arith.constant 2 : index
    %res = arith.muli %n, %two : index
    func.return %res : index
  }
}
"""

    ctx = Context(True)
    for dialect_name, dialect_factory in get_all_dialects().items():
        ctx.register_dialect(dialect_name, dialect_factory)
    parser = Parser(ctx, text)
    module = parser.parse_module()

    expected_res = {
        "arith.constant": 1,
        "arith.muli": 1,
        "builtin.module": 1,
        "func.func": 1,
        "func.return": 1,
    }

    res = count_number_of_operations(module)
    assert res == expected_res


def test_get_diff_operation_count():
    # get input module
    input_text = """builtin.module {
  func.func @hello(%n : index) -> index {
    %two = arith.constant 2 : index
    %res = arith.muli %n, %two : index
    func.return %res : index
  }
}
"""

    ctx = Context(True)
    for dialect_name, dialect_factory in get_all_dialects().items():
        ctx.register_dialect(dialect_name, dialect_factory)
    parser = Parser(ctx, input_text)
    input_module = parser.parse_module()

    # get output module
    output_text = """builtin.module {
  func.func @hello(%n : index) -> index {
    %two = riscv.li 2 : !riscv.reg
    %two_1 = builtin.unrealized_conversion_cast %two : !riscv.reg to index
    %res = builtin.unrealized_conversion_cast %n : index to !riscv.reg
    %res_1 = builtin.unrealized_conversion_cast %two_1 : index to !riscv.reg
    %res_2 = riscv.mul %res, %res_1 : (!riscv.reg, !riscv.reg) -> !riscv.reg
    %res_3 = builtin.unrealized_conversion_cast %res_2 : !riscv.reg to index
    func.return %res_3 : index
  }
}
"""
    parser = Parser(ctx.clone(), output_text)
    output_module = parser.parse_module()

    expected_diff_res: tuple[tuple[str, int, str], ...] = (
        ("arith.constant", 0, "-1"),
        ("arith.muli", 0, "-1"),
        ("builtin.module", 1, "="),
        ("builtin.unrealized_conversion_cast", 4, "+4"),
        ("func.func", 1, "="),
        ("func.return", 1, "="),
        ("riscv.li", 1, "+1"),
        ("riscv.mul", 1, "+1"),
    )

    assert expected_diff_res == get_diff_operation_count(
        tuple(count_number_of_operations(input_module).items()),
        tuple(count_number_of_operations(output_module).items()),
    )



import re
from collections.abc import Sequence
from io import StringIO

import pytest

from xdsl.builder import ImplicitBuilder
from xdsl.context import Context
from xdsl.dialects import test
from xdsl.dialects.arith import AddiOp, Arith, ConstantOp, MuliOp
from xdsl.dialects.builtin import (
    Builtin,
    IndexType,
    IntegerAttr,
    IntegerType,
    ModuleOp,
    StringAttr,
    UnitAttr,
    i32,
    i64,
)
from xdsl.ir import Block, Operation, SSAValue
from xdsl.irdl import BaseAttr
from xdsl.parser import Parser
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriterListener,
    PatternRewriteWalker,
    RewritePattern,
    TypeConversionPattern,
    attr_constr_rewrite_pattern,
    attr_type_rewrite_pattern,
    op_type_rewrite_pattern,
)
from xdsl.printer import Printer
from xdsl.rewriter import BlockInsertPoint, InsertPoint


def rewrite_and_compare(
    prog: str,
    expected_prog: str,
    walker: PatternRewriteWalker,
    *,
    op_inserted: int = 0,
    op_removed: int = 0,
    op_modified: int = 0,
    op_replaced: int = 0,
    block_created: int = 0,
    expect_rewrite: bool = True,
):
    ctx = Context(allow_unregistered=True)
    ctx.load_dialect(Builtin)
    ctx.load_dialect(Arith)
    ctx.load_dialect(test.Test)

    num_op_inserted = 0
    num_op_removed = 0
    num_op_modified = 0
    num_op_replaced = 0
    num_block_created = 0

    def op_inserted_handler(op: Operation):
        nonlocal num_op_inserted
        num_op_inserted += 1

    def op_removed_handler(op: Operation):
        nonlocal num_op_removed
        num_op_removed += 1

    def op_modified_handler(op: Operation):
        nonlocal num_op_modified
        num_op_modified += 1

    def op_replaced_handler(op: Operation, values: Sequence[SSAValue | None]):
        nonlocal num_op_replaced
        num_op_replaced += 1

    def block_created_handler(block: Block):
        nonlocal num_block_created
        num_block_created += 1

    listener = PatternRewriterListener()
    listener.operation_insertion_handler = [op_inserted_handler]
    listener.operation_removal_handler = [op_removed_handler]
    listener.operation_modification_handler = [op_modified_handler]
    listener.operation_replacement_handler = [op_replaced_handler]
    listener.block_creation_handler = [block_created_handler]

    parser = Parser(ctx, prog)
    module = parser.parse_module()

    walker.listener = listener
    did_rewrite = walker.rewrite_module(module)

    file = StringIO()
    printer = Printer(stream=file, print_generic_format=True)
    printer.print(module)

    assert file.getvalue().strip() == expected_prog.strip()

    assert num_op_inserted == op_inserted
    assert num_op_removed == op_removed
    assert num_op_modified == op_modified
    assert num_op_replaced == op_replaced
    assert num_block_created == block_created
    assert did_rewrite == expect_rewrite


def test_non_recursive_rewrite():
    """Test a simple non-recursive rewrite"""

    prog = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 42 : i32}> : () -> i32
  %1 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()"""

    expected = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 43 : i32}> : () -> i32
  %1 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()"""

    class RewriteConst(RewritePattern):
        def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
            if isinstance(op, ConstantOp):
                new_constant = ConstantOp.from_int_and_width(43, i32)
                rewriter.replace_matched_op([new_constant])

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(RewriteConst(), apply_recursively=False),
        op_inserted=1,
        op_removed=1,
        op_replaced=1,
        op_modified=2,
    )


def test_non_recursive_rewrite_reversed():
    """Test a simple non-recursive rewrite with reverse walk order."""

    prog = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 42 : i32}> : () -> i32
  %1 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()"""

    expected = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 43 : i32}> : () -> i32
  %1 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()"""

    class RewriteConst(RewritePattern):
        def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
            if isinstance(op, ConstantOp):
                new_constant = ConstantOp.from_int_and_width(43, i32)
                rewriter.replace_matched_op([new_constant])

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(
            RewriteConst(), apply_recursively=False, walk_reverse=True
        ),
        op_inserted=1,
        op_removed=1,
        op_replaced=1,
        op_modified=2,
    )


def test_op_type_rewrite_pattern_method_decorator():
    """Test op_type_rewrite_pattern decorator on methods."""

    prog = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 42 : i32}> : () -> i32
  %1 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()"""

    expected = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 43 : i32}> : () -> i32
  %1 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()"""

    class RewriteConst(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, __op__: ConstantOp, rewriter: PatternRewriter):
            rewriter.replace_matched_op(ConstantOp.from_int_and_width(43, i32))

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(RewriteConst(), apply_recursively=False),
        op_inserted=1,
        op_removed=1,
        op_replaced=1,
        op_modified=2,
    )


def test_op_type_rewrite_pattern_union_type():
    """Test op_type_rewrite_pattern decorator on static functions."""

    prog = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 42 : i32}> : () -> i32
  %1 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
  %2 = "test"(%0, %1) : (i32, i32) -> i32
}) : () -> ()"""

    expected = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 42 : i32}> : () -> i32
  %1 = "arith.constant"() <{value = 42 : i32}> : () -> i32
  %2 = "test"(%0, %1) : (i32, i32) -> i32
}) : () -> ()"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(
            self, __op__: ConstantOp | AddiOp, rewriter: PatternRewriter
        ):
            rewriter.replace_matched_op(ConstantOp.from_int_and_width(42, i32))

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
        op_inserted=2,
        op_removed=2,
        op_replaced=2,
        op_modified=4,
    )


def test_recursive_rewriter():
    """Test recursive walks on operations created by rewrites."""

    prog = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 5 : i32}> : () -> i32
}) : () -> ()"""

    expected = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 1 : i32}> : () -> i32
  %1 = "arith.constant"() <{value = 1 : i32}> : () -> i32
  %2 = "arith.addi"(%0, %1) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
  %3 = "arith.constant"() <{value = 1 : i32}> : () -> i32
  %4 = "arith.addi"(%2, %3) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
  %5 = "arith.constant"() <{value = 1 : i32}> : () -> i32
  %6 = "arith.addi"(%4, %5) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
  %7 = "arith.constant"() <{value = 1 : i32}> : () -> i32
  %8 = "arith.addi"(%6, %7) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: ConstantOp, rewriter: PatternRewriter):
            if not isinstance(op_val := op.value, IntegerAttr):
                return
            val = op_val.value.data
            if val == 0 or val == 1:
                return
            constant_op = ConstantOp(IntegerAttr.from_int_and_width(val - 1, 32), i32)
            constant_one = ConstantOp(IntegerAttr.from_int_and_width(1, 32), i32)
            add_op = AddiOp(constant_op, constant_one)
            rewriter.replace_matched_op([constant_op, constant_one, add_op])

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=True),
        op_inserted=12,
        op_removed=4,
        op_replaced=4,
        op_modified=3,
    )


def test_recursive_rewriter_reversed():
    """Test recursive walks on operations created by rewrites, in reverse walk order."""

    prog = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 5 : i32}> : () -> i32
}) : () -> ()"""

    expected = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 1 : i32}> : () -> i32
  %1 = "arith.constant"() <{value = 1 : i32}> : () -> i32
  %2 = "arith.addi"(%0, %1) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
  %3 = "arith.constant"() <{value = 1 : i32}> : () -> i32
  %4 = "arith.addi"(%2, %3) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
  %5 = "arith.constant"() <{value = 1 : i32}> : () -> i32
  %6 = "arith.addi"(%4, %5) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
  %7 = "arith.constant"() <{value = 1 : i32}> : () -> i32
  %8 = "arith.addi"(%6, %7) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: ConstantOp, rewriter: PatternRewriter):
            if not isinstance(op_val := op.value, IntegerAttr):
                return
            val = op_val.value.data
            if val == 0 or val == 1:
                return
            constant_op = ConstantOp(IntegerAttr.from_int_and_width(val - 1, 32), i32)
            constant_one = ConstantOp(IntegerAttr.from_int_and_width(1, 32), i32)
            add_op = AddiOp(constant_op, constant_one)
            rewriter.replace_matched_op([constant_op, constant_one, add_op])

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=True, walk_reverse=True),
        op_inserted=12,
        op_removed=4,
        op_replaced=4,
        op_modified=3,
    )


def test_greedy_rewrite_pattern_applier():
    """Test GreedyRewritePatternApplier."""

    prog = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 42 : i32}> : () -> i32
  %1 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()"""

    expected = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 43 : i32}> : () -> i32
  %1 = "arith.muli"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()"""

    class ConstantRewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, __op__: ConstantOp, rewriter: PatternRewriter):
            rewriter.replace_matched_op([ConstantOp.from_int_and_width(43, i32)])

    class AddiRewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: AddiOp, rewriter: PatternRewriter):
            rewriter.replace_matched_op([MuliOp(op.lhs, op.rhs)])

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(
            GreedyRewritePatternApplier([ConstantRewrite(), AddiRewrite()]),
            apply_recursively=False,
        ),
        op_inserted=2,
        op_removed=2,
        op_replaced=2,
        op_modified=2,
    )


def test_insert_op_before_matched_op():
    """Test rewrites where operations are inserted before the matched operation."""

    prog = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 5 : i32}> : () -> i32
}) : () -> ()"""

    expected = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 42 : i32}> : () -> i32
  %1 = "arith.constant"() <{value = 5 : i32}> : () -> i32
}) : () -> ()"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, __cst__: ConstantOp, rewriter: PatternRewriter):
            new_cst = ConstantOp.from_int_and_width(42, i32)

            rewriter.insert_op_before_matched_op(new_cst)

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
        op_inserted=1,
    )


def test_insert_op_at_start():
    """Test rewrites where operations are inserted with a given position."""

    prog = """"builtin.module"() ({
  "test.op"() ({
    %0 = "arith.constant"() <{value = 5 : i32}> : () -> i32
  }) : () -> ()
}) : () -> ()"""

    expected = """"builtin.module"() ({
  "test.op"() ({
    %0 = "arith.constant"() <{value = 42 : i32}> : () -> i32
    %1 = "arith.constant"() <{value = 5 : i32}> : () -> i32
  }) : () -> ()
}) : () -> ()"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: test.TestOp, rewriter: PatternRewriter):
            new_cst = ConstantOp.from_int_and_width(42, i32)

            rewriter.insert_op(new_cst, InsertPoint.at_start(op.regions[0].blocks[0]))

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
        op_inserted=1,
    )


def test_insert_op_before():
    """Test rewrites where operations are inserted before a given operation."""

    prog = """"builtin.module"() ({
  "test.op"() ({
    %0 = "arith.constant"() <{value = 5 : i32}> : () -> i32
  }) : () -> ()
}) : () -> ()"""

    expected = """"builtin.module"() ({
  "test.op"() ({
    %0 = "arith.constant"() <{value = 42 : i32}> : () -> i32
    %1 = "arith.constant"() <{value = 5 : i32}> : () -> i32
  }) : () -> ()
}) : () -> ()"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: test.TestOp, rewriter: PatternRewriter):
            new_cst = ConstantOp.from_int_and_width(42, i32)

            first_op = op.regions[0].block.ops.first
            assert first_op is not None
            rewriter.insert_op(new_cst, InsertPoint.before(first_op))

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
        op_inserted=1,
    )


def test_insert_op_after():
    """Test rewrites where operations are inserted after a given operation."""

    prog = """"builtin.module"() ({
  "test.op"() ({
    %0 = "arith.constant"() <{value = 5 : i32}> : () -> i32
  }) : () -> ()
}) : () -> ()"""

    expected = """"builtin.module"() ({
  "test.op"() ({
    %0 = "arith.constant"() <{value = 5 : i32}> : () -> i32
    %1 = "arith.constant"() <{value = 42 : i32}> : () -> i32
  }) : () -> ()
}) : () -> ()"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: test.TestOp, rewriter: PatternRewriter):
            new_cst = ConstantOp.from_int_and_width(42, i32)

            first_op = op.regions[0].block.ops.first
            assert first_op is not None
            rewriter.insert_op(new_cst, InsertPoint.after(first_op))

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
        op_inserted=1,
    )


def test_insert_op_after_matched_op():
    """Test rewrites where operations are inserted after a given operation."""

    prog = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 5 : i32}> : () -> i32
}) : () -> ()"""

    expected = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 5 : i32}> : () -> i32
  %1 = "arith.constant"() <{value = 42 : i32}> : () -> i32
}) : () -> ()"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, __cst__: ConstantOp, rewriter: PatternRewriter):
            new_cst = ConstantOp.from_int_and_width(42, i32)

            rewriter.insert_op_after_matched_op(new_cst)

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
        op_inserted=1,
    )


def test_insert_op_after_matched_op_reversed():
    """Test rewrites where operations are inserted after a given operation."""

    prog = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 5 : i32}> : () -> i32
}) : () -> ()"""

    expected = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 5 : i32}> : () -> i32
  %1 = "arith.constant"() <{value = 42 : i32}> : () -> i32
}) : () -> ()"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, __cst__: ConstantOp, rewriter: PatternRewriter):
            new_cst = ConstantOp.from_int_and_width(42, i32)

            rewriter.insert_op_after_matched_op(new_cst)

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False, walk_reverse=True),
        op_inserted=1,
    )


def test_operation_deletion():
    """Test rewrites where SSA values are deleted."""

    prog = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 5 : i32}> : () -> i32
}) : () -> ()"""

    expected = """"builtin.module"() ({
^0:
}) : () -> ()"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, __op__: ConstantOp, rewriter: PatternRewriter):
            rewriter.erase_matched_op()

    rewrite_and_compare(prog, expected, PatternRewriteWalker(Rewrite()), op_removed=1)


def test_operation_deletion_reversed():
    """
    Test rewrites where SSA values are deleted.
    They have to be deleted in order for the rewrite to not fail.
    """

    prog = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 5 : i32}> : () -> i32
  %1 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()"""

    expected = """"builtin.module"() ({
^0:
}) : () -> ()"""

    class EraseAll(RewritePattern):
        def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
            if not isinstance(op, ModuleOp):
                rewriter.erase_matched_op()

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(EraseAll(), walk_reverse=True),
        op_removed=2,
    )


def test_operation_deletion_failure():
    """Test rewrites where SSA values are deleted with still uses."""

    ctx = Context()
    ctx.load_dialect(Builtin)
    ctx.load_dialect(Arith)

    prog = """"builtin.module"() ({
  %0 = "arith.constant"() <{value = 5 : i32}> : () -> i32
  %1 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, __op__: ConstantOp, rewriter: PatternRewriter):
            rewriter.erase_matched_op()

    parser = Parser(ctx, prog)
    module = parser.parse_module()
    walker = PatternRewriteWalker(Rewrite())

    # Check that the rewrite fails
    with pytest.raises(
        Exception,
        match="Attempting to delete SSA value that still has uses of result of operation",
    ):
        walker.rewrite_module(module)


def test_delete_inner_op():
    """Test rewrites where an operation inside a region of the matched op is deleted."""

    prog = """"builtin.module"() ({
  "test.op"() ({
    %0 = "arith.constant"() <{value = 5 : i32}> : () -> i32
  }) : () -> ()
}) : () -> ()"""

    expected = """"builtin.module"() ({
  "test.op"() ({
  ^0:
  }) : () -> ()
}) : () -> ()"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: test.TestOp, rewriter: PatternRewriter):
            first_op = op.regions[0].block.ops.first

            assert first_op is not None
            rewriter.erase_op(first_op)

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
        op_removed=1,
    )


def test_replace_inner_op():
    """Test rewrites where an operation inside a region of the matched op is deleted."""

    prog = """"builtin.module"() ({
  "test.op"() ({
    %0 = "arith.constant"() <{value = 5 : i32}> : () -> i32
  }) : () -> ()
}) : () -> ()"""

    expected = """"builtin.module"() ({
  "test.op"() ({
    %0 = "arith.constant"() <{value = 42 : i32}> : () -> i32
  }) : () -> ()
}) : () -> ()"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: test.TestOp, rewriter: PatternRewriter):
            first_op = op.regions[0].block.ops.first

            assert first_op is not None
            rewriter.replace_op(first_op, [ConstantOp.from_int_and_width(42, i32)])

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
        op_inserted=1,
        op_removed=1,
        op_replaced=1,
    )


def test_block_argument_type_change():
    """Test the modification of a block argument type."""

    prog = """\
"builtin.module"() ({
  %0 = "test.op"() ({
  ^0(%1 : !test.type<"int">):
    %2 = "test.op"() : () -> !test.type<"int">
  }) : () -> !test.type<"int">
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() ({
  ^0(%1 : i64):
    %2 = "test.op"() : () -> !test.type<"int">
  }) : () -> !test.type<"int">
}) : () -> ()
"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, matched_op: test.TestOp, rewriter: PatternRewriter):
            if matched_op.regs and matched_op.regs[0].blocks:
                rewriter.modify_value_type(matched_op.regs[0].blocks[0].args[0], i64)

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
        op_modified=1,
    )


def test_block_argument_erasure():
    """Test the erasure of a block argument."""

    prog = """\
"builtin.module"() ({
  %0 = "test.op"() ({
  ^0(%1 : !test.type<"int">):
    %2 = "test.op"() : () -> !test.type<"int">
  }, {
  ^0:
  }) : () -> !test.type<"int">
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() ({
    %1 = "test.op"() : () -> !test.type<"int">
  }, {
  ^0:
  }) : () -> !test.type<"int">
}) : () -> ()
"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, matched_op: test.TestOp, rewriter: PatternRewriter):
            if matched_op.regs and matched_op.regs[0].blocks:
                rewriter.erase_block_argument(matched_op.regs[0].blocks[0].args[0])

    rewrite_and_compare(
        prog, expected, PatternRewriteWalker(Rewrite(), apply_recursively=False)
    )


def test_block_argument_insertion():
    """Test the insertion of a block argument."""

    prog = """\
"builtin.module"() ({
  %0 = "test.op"() ({
    %1 = "test.op"() : () -> !test.type<"int">
  }, {
  ^0:
  }) : () -> !test.type<"int">
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() ({
  ^0(%1 : !test.type<"int">):
    %2 = "test.op"() : () -> !test.type<"int">
  }, {
  ^1:
  }) : () -> !test.type<"int">
}) : () -> ()
"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, matched_op: test.TestOp, rewriter: PatternRewriter):
            if matched_op.regs and matched_op.regs[0].blocks:
                rewriter.insert_block_argument(
                    matched_op.regs[0].blocks[0], 0, test.TestType("int")
                )

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
    )


def test_inline_block_before_matched_op():
    """Test the inlining of a block before the matched operation."""

    prog = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> !test.type<"int">
  %1 = "test.op"() ({
  ^0:
    %2 = "test.op"() : () -> !test.type<"int">
  }, {
  ^1:
  }) : () -> !test.type<"int">
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> !test.type<"int">
  %1 = "test.op"() : () -> !test.type<"int">
  %2 = "test.op"() ({
  }, {
  ^0:
  }) : () -> !test.type<"int">
}) : () -> ()
"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, matched_op: test.TestOp, rewriter: PatternRewriter):
            if matched_op.regs and matched_op.regs[0].blocks:
                rewriter.inline_block_before_matched_op(matched_op.regs[0].blocks[0])

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
    )


def test_inline_block_before():
    """Test the inlining of a block before an operation."""

    prog = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> !test.type<"int">
  %1 = "test.op"() ({
    %1 = "test.op"() ({
      %1 = "test.op"() : () -> !test.type<"int">
    }, {
    ^1:
    }) : () -> !test.type<"int">
  }, {
  ^2:
  }) : () -> !test.type<"int">
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> !test.type<"int">
  %1 = "test.op"() ({
    %2 = "test.op"() : () -> !test.type<"int">
    %3 = "test.op"() ({
    }, {
    ^0:
    }) : () -> !test.type<"int">
  }, {
  ^1:
  }) : () -> !test.type<"int">
}) : () -> ()
"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, matched_op: test.TestOp, rewriter: PatternRewriter):
            if matched_op.regs and matched_op.regs[0].blocks:
                first_op = matched_op.regs[0].blocks[0].first_op

                if isinstance(first_op, test.TestOp):
                    inner_block = first_op.regs[0].blocks[0]
                    rewriter.inline_block(inner_block, InsertPoint.before(first_op))

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
    )


def test_inline_block_at_before_when_op_is_matched_op():
    """Test the inlining of a block before an operation, being the matched one."""

    prog = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> !test.type<"int">
  %1 = "test.op"() ({
  ^0:
    %2 = "test.op"() : () -> !test.type<"int">
  }, {
  ^1:
  }) : () -> !test.type<"int">
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> !test.type<"int">
  %1 = "test.op"() : () -> !test.type<"int">
  %2 = "test.op"() ({
  }, {
  ^0:
  }) : () -> !test.type<"int">
}) : () -> ()
"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, matched_op: test.TestOp, rewriter: PatternRewriter):
            if matched_op.regs and matched_op.regs[0].blocks:
                rewriter.inline_block(
                    matched_op.regs[0].blocks[0], InsertPoint.before(matched_op)
                )

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
    )


def test_inline_block_before_with_args():
    """Test the inlining of a block before an operation."""

    prog = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> !test.type<"int">
  %1 = "test.op"() ({
  ^0(%arg0 : !test.type<"int">):
    %1 = "test.op"() ({
    ^1(%arg1 : !test.type<"int">):
      %1 = "test.op"(%arg1) : (!test.type<"int">) -> !test.type<"int">
    }, {
    ^2:
    }) : () -> !test.type<"int">
  }, {
  ^3:
  }) : () -> !test.type<"int">
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> !test.type<"int">
  %1 = "test.op"() ({
  ^0(%arg0 : !test.type<"int">):
    %2 = "test.op"(%arg0) : (!test.type<"int">) -> !test.type<"int">
    %3 = "test.op"() ({
    }, {
    ^1:
    }) : () -> !test.type<"int">
  }, {
  ^2:
  }) : () -> !test.type<"int">
}) : () -> ()
"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, matched_op: test.TestOp, rewriter: PatternRewriter):
            if matched_op.regs and matched_op.regs[0].blocks:
                outer_block = matched_op.regs[0].blocks[0]
                first_op = outer_block.first_op

                if isinstance(first_op, test.TestOp):
                    inner_block = first_op.regs[0].blocks[0]
                    rewriter.inline_block(
                        inner_block, InsertPoint.before(first_op), outer_block.args
                    )

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
    )


def test_inline_block_after():
    """Test the inlining of a block after an operation."""

    prog = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> !test.type<"int">
  %1 = "test.op"() ({
    %1 = "test.op"() ({
      %1 = "test.op"() : () -> !test.type<"int">
    }, {
      ^1:
    }) : () -> !test.type<"int">
  }, {
  ^2:
  }) : () -> !test.type<"int">
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> !test.type<"int">
  %1 = "test.op"() ({
    %2 = "test.op"() ({
    }, {
    ^0:
    }) : () -> !test.type<"int">
    %3 = "test.op"() : () -> !test.type<"int">
  }, {
  ^1:
  }) : () -> !test.type<"int">
}) : () -> ()
"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, matched_op: test.TestOp, rewriter: PatternRewriter):
            if matched_op.regs and matched_op.regs[0].blocks:
                first_op = matched_op.regs[0].blocks[0].first_op

                if first_op is not None and isinstance(first_op, test.TestOp):
                    if first_op.regs and first_op.regs[0].blocks:
                        inner_block = first_op.regs[0].blocks[0]
                        rewriter.inline_block(inner_block, InsertPoint.after(first_op))

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
    )


def test_inline_block_after_matched():
    """Test the inlining of a block after an operation."""

    prog = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> !test.type<"int">
  %1 = "test.op"() ({
    %1 = "test.op"() ({
      %1 = "test.op"() : () -> !test.type<"int">
    }, {
    ^1:
    }) : () -> !test.type<"int">
  }, {
  ^2:
  }) : () -> !test.type<"int">
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> !test.type<"int">
  %1 = "test.op"() ({
    %2 = "test.op"() ({
    }, {
    ^0:
    }) : () -> !test.type<"int">
  }, {
  ^1:
  }) : () -> !test.type<"int">
  %3 = "test.op"() : () -> !test.type<"int">
}) : () -> ()
"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, matched_op: test.TestOp, rewriter: PatternRewriter):
            if matched_op.regs and matched_op.regs[0].blocks:
                first_op = matched_op.regs[0].block.first_op

                if first_op is not None and isinstance(first_op, test.TestOp):
                    if first_op.regs and first_op.regs[0].blocks:
                        inner_block = first_op.regs[0].blocks[0]
                        rewriter.inline_block(
                            inner_block, InsertPoint.after(matched_op)
                        )

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
    )


def test_move_region_contents_to_new_regions():
    """Test moving a region outside of a region."""

    prog = """\
"builtin.module"() ({
  "test.op"() ({
    %0 = "test.op"() : () -> !test.type<"int">
    %1 = "test.op"() ({
    ^0:
      %2 = "test.op"() : () -> !test.type<"int">
    }) : () -> !test.type<"int">
  }) : () -> ()
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  "test.op"() ({
    %0 = "test.op"() : () -> !test.type<"int">
    %1 = "test.op"() ({
    }) : () -> !test.type<"int">
    %2 = "test.op"() ({
      %3 = "test.op"() : () -> !test.type<"int">
    }) : () -> !test.type<"int">
  }) : () -> ()
}) : () -> ()
"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: test.TestOp, rewriter: PatternRewriter):
            # Match the toplevel test.op
            if not isinstance(op.parent_op(), ModuleOp):
                return
            ops_iter = iter(op.regions[0].block.ops)

            _ = next(ops_iter)  # skip first op
            old_op = next(ops_iter)
            assert isinstance(old_op, test.TestOp)
            new_region = rewriter.move_region_contents_to_new_regions(old_op.regions[0])
            res_types = old_op.result_types
            new_op = test.TestOp.create(result_types=res_types, regions=[new_region])
            rewriter.insert_op(new_op, InsertPoint.after(old_op))

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
        op_inserted=1,
    )


def test_insert_same_block():
    """Test rewriter on ops without results"""
    prog = """\
"builtin.module"() ({
  %0 = "test.op"() {label = "a"} : () -> i32
  %1 = "test.op"() {label = "b"} : () -> i32
  %2 = "test.op"() {label = "c"} : () -> i32
  "func.return"() : () -> ()
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() {label = "alloc"} : () -> i32
  %1 = "test.op"() {label = "a"} : () -> i32
  "test.op"(%0) {label = "init"} : (i32) -> ()
  %2 = "test.op"() {label = "c"} : () -> i32
  "test.op"(%0) {label = "dealloc"} : (i32) -> ()
  "func.return"() : () -> ()
}) : () -> ()
"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: test.TestOp, rewriter: PatternRewriter):
            if op.attributes["label"] != StringAttr("b"):
                return

            block = op.parent

            if block is None:
                return

            last_op = block.last_op
            assert last_op is not None

            alloc = test.TestOp.build(
                operands=((),),
                attributes={"label": StringAttr("alloc")},
                regions=((),),
                result_types=((i32,),),
            )
            init = test.TestOp.build(
                operands=(alloc.res,),
                attributes={"label": StringAttr("init")},
                regions=((),),
                result_types=((),),
            )
            dealloc = test.TestOp.build(
                operands=(alloc.res,),
                attributes={"label": StringAttr("dealloc")},
                regions=((),),
                result_types=((),),
            )

            # Allocate before first use
            rewriter.insert_op(alloc, InsertPoint.at_start(block))
            # Deallocate after last use
            rewriter.insert_op(dealloc, InsertPoint.before(last_op))
            # Init instead of creating, and replace result with allocated value
            rewriter.replace_matched_op(init, alloc.res)

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
        op_inserted=3,
        op_removed=1,
        op_replaced=1,
    )


def test_inline_region_before():
    prog = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> i32
^0:
  "test.op"() ({
    ^1:
    %1 = "test.op"() : () -> f32
    ^2:
    %2 = "test.op"() : () -> f64
  }) {label = "а"} : () -> ()
  %2 = "test.op"() : () -> i64
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> i32
^0:
  %1 = "test.op"() : () -> f32
^1:
  %2 = "test.op"() : () -> f64
^2:
  %3 = "test.op"() : () -> i64
}) : () -> ()
"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: test.TestOp, rewriter: PatternRewriter):
            if op.attributes.get("label") != StringAttr("а"):
                return
            if op.parent is None:
                return

            rewriter.inline_region(op.regions[0], BlockInsertPoint.before(op.parent))
            rewriter.erase_matched_op()

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
        op_inserted=0,
        op_removed=1,
    )


def test_inline_region_after():
    prog = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> i32
  "test.op"() ({
    ^1:
    %1 = "test.op"() : () -> f32
    ^2:
    %2 = "test.op"() : () -> f64
  }) {label = "а"} : () -> ()
^0:
  %2 = "test.op"() : () -> i64
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> i32
^0:
  %1 = "test.op"() : () -> f32
^1:
  %2 = "test.op"() : () -> f64
^2:
  %3 = "test.op"() : () -> i64
}) : () -> ()
"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: test.TestOp, rewriter: PatternRewriter):
            if op.attributes.get("label") != StringAttr("а"):
                return
            if op.parent is None:
                return

            rewriter.inline_region(op.regions[0], BlockInsertPoint.after(op.parent))
            rewriter.erase_matched_op()

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
        op_inserted=0,
        op_removed=1,
    )


def test_inline_region_at_start():
    prog = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> i32
^0:
  "test.op"() ({
    ^1:
    %1 = "test.op"() : () -> f32
    ^2:
    %2 = "test.op"() : () -> f64
  }) {label = "а"} : () -> ()
  %2 = "test.op"() : () -> i64
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> f32
^0:
  %1 = "test.op"() : () -> f64
^1:
  %2 = "test.op"() : () -> i32
^2:
  %3 = "test.op"() : () -> i64
}) : () -> ()
"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: test.TestOp, rewriter: PatternRewriter):
            if op.attributes.get("label") != StringAttr("а"):
                return
            parent_region = op.parent_region()
            if parent_region is None:
                return

            rewriter.inline_region(
                op.regions[0], BlockInsertPoint.at_start(parent_region)
            )
            rewriter.erase_matched_op()

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
        op_inserted=0,
        op_removed=1,
    )


def test_inline_region_at_end():
    prog = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> i32
^0:
  "test.op"() ({
    ^1:
    %1 = "test.op"() : () -> f32
    ^2:
    %2 = "test.op"() : () -> f64
  }) {label = "а"} : () -> ()
  %2 = "test.op"() : () -> i64
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> i32
^0:
  %1 = "test.op"() : () -> i64
^1:
  %2 = "test.op"() : () -> f32
^2:
  %3 = "test.op"() : () -> f64
}) : () -> ()
"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: test.TestOp, rewriter: PatternRewriter):
            if op.attributes.get("label") != StringAttr("а"):
                return
            parent_region = op.parent_region()
            if parent_region is None:
                return

            rewriter.inline_region(
                op.regions[0], BlockInsertPoint.at_end(parent_region)
            )
            rewriter.erase_matched_op()

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
        op_inserted=0,
        op_removed=1,
    )


def test_erased_ssavalue():
    prog = """\
builtin.module {
  "test.op"() ({
    %0 = "test.op"() : () -> i32
    "test.op"(%0) : (i32) -> ()
  }) : () -> ()
}
  """

    expected = """\
"builtin.module"() ({
  "test.op"() ({
  ^0:
  }) : () -> ()
}) : () -> ()
"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: test.TestOp, rewriter: PatternRewriter):
            if op.results or op.operands:
                rewriter.erase_matched_op(safe_erase=False)

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=True),
        op_removed=2,
    )


def test_pattern_rewriter_as_op_builder():
    """Test that the PatternRewriter works as an OpBuilder."""
    prog = """
"builtin.module"() ({
  "test.op"() : () -> ()
  "test.op"() {nomatch} : () -> ()
  "test.op"() : () -> ()
}) : () -> ()"""
    expected = """
"builtin.module"() ({
  "test.op"() {inserted} : () -> ()
  "test.op"() {replaced} : () -> ()
  "test.op"() {nomatch} : () -> ()
  "test.op"() {inserted} : () -> ()
  "test.op"() {replaced} : () -> ()
}) : () -> ()"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: test.TestOp, rewriter: PatternRewriter):
            if "nomatch" in op.attributes:
                return
            with ImplicitBuilder(rewriter):
                test.TestOp.create(attributes={"inserted": UnitAttr()})
            rewriter.replace_matched_op(
                test.TestOp.create(attributes={"replaced": UnitAttr()})
            )

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
        op_inserted=4,
        op_removed=2,
        op_replaced=2,
    )


def test_type_conversion():
    """Test rewriter on ops without results"""
    prog = """\
"builtin.module"() ({
  "func.func"() ({
  ^0(%arg : i32):
  }) : () -> ()
  %0 = "test.op"() {nested = memref<*xi32>} : () -> i32
  %1 = "test.op"() {type = () -> memref<*xi32>} : () -> f32
  %2 = "test.op"() <{prop = memref<*xi32>}> : () -> f32
  %3 = "test.op"(%0, %1) : (i32, f32) -> memref<*xi32>
  %4 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
  "func.return"() : () -> ()
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  "func.func"() ({
  ^0(%arg : index):
  }) : () -> ()
  %0 = "test.op"() {nested = memref<*xindex>} : () -> index
  %1 = "test.op"() {type = () -> memref<*xindex>} : () -> f32
  %2 = "test.op"() <{prop = memref<*xindex>}> : () -> f32
  %3 = "test.op"(%0, %1) : (index, f32) -> memref<*xindex>
  %4 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (index, index) -> index
  "func.return"() : () -> ()
}) : () -> ()
"""

    class Rewrite(TypeConversionPattern):
        @attr_type_rewrite_pattern
        def convert_type(self, typ: IntegerType) -> IndexType:
            return IndexType()

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(recursive=True), apply_recursively=False),
        op_inserted=5,
        op_removed=5,
        op_replaced=5,
        op_modified=5,
    )
    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(recursive=True), apply_recursively=True),
        op_inserted=5,
        op_removed=5,
        op_replaced=5,
        op_modified=5,
    )
    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(
            Rewrite(recursive=True), apply_recursively=False, walk_reverse=True
        ),
        op_inserted=5,
        op_removed=5,
        op_replaced=5,
        op_modified=5,
    )

    non_rec_expected = """\
"builtin.module"() ({
  "func.func"() ({
  ^0(%arg : index):
  }) : () -> ()
  %0 = "test.op"() {nested = memref<*xi32>} : () -> index
  %1 = "test.op"() {type = () -> memref<*xi32>} : () -> f32
  %2 = "test.op"() <{prop = memref<*xi32>}> : () -> f32
  %3 = "test.op"(%0, %1) : (index, f32) -> memref<*xi32>
  %4 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (index, index) -> index
  "func.return"() : () -> ()
}) : () -> ()
"""

    rewrite_and_compare(
        prog,
        non_rec_expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
        op_inserted=2,
        op_removed=2,
        op_replaced=2,
        op_modified=4,
    )
    rewrite_and_compare(
        prog,
        non_rec_expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=True),
        op_inserted=2,
        op_removed=2,
        op_replaced=2,
        op_modified=4,
    )
    rewrite_and_compare(
        prog,
        non_rec_expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False, walk_reverse=True),
        op_inserted=2,
        op_removed=2,
        op_replaced=2,
        op_modified=4,
    )

    expected = """\
"builtin.module"() ({
  "func.func"() ({
  ^0(%arg : i32):
  }) : () -> ()
  %0 = "test.op"() {nested = memref<*xindex>} : () -> index
  %1 = "test.op"() {type = () -> memref<*xindex>} : () -> f32
  %2 = "test.op"() <{prop = memref<*xindex>}> : () -> f32
  %3 = "test.op"(%0, %1) : (index, f32) -> memref<*xindex>
  %4 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (index, index) -> i32
  "func.return"() : () -> ()
}) : () -> ()
"""

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(
            Rewrite(ops=(test.TestOp,), recursive=True), apply_recursively=False
        ),
        op_inserted=4,
        op_removed=4,
        op_replaced=4,
        op_modified=4,
    )
    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(
            Rewrite(ops=(test.TestOp,), recursive=True), apply_recursively=True
        ),
        op_inserted=4,
        op_removed=4,
        op_replaced=4,
        op_modified=4,
    )
    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(
            Rewrite(ops=(test.TestOp,), recursive=True),
            apply_recursively=False,
            walk_reverse=True,
        ),
        op_inserted=4,
        op_removed=4,
        op_replaced=4,
        op_modified=4,
    )

    class RewriteMaybe(TypeConversionPattern):
        @attr_type_rewrite_pattern
        def convert_type(self, typ: IntegerType) -> IndexType | None:
            return IndexType() if typ.width.data >= 8 else None

    prog = """\
"builtin.module"() ({
  "func.func"() ({
  ^0(%arg : i6):
  }) : () -> ()
  %0 = "test.op"() {nested = memref<*xi4>} : () -> i6
  %1 = "test.op"() {type = () -> memref<*xi32>} : () -> f32
  %2 = "test.op"() <{prop = memref<*xi4>}> : () -> i32
  %3 = "test.op"(%0, %1) : (i6, f32) -> memref<*xi32>
  %4 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i6, i6) -> i6
  "func.return"() : () -> ()
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  "func.func"() ({
  ^0(%arg : i6):
  }) : () -> ()
  %0 = "test.op"() {nested = memref<*xi4>} : () -> i6
  %1 = "test.op"() {type = () -> memref<*xindex>} : () -> f32
  %2 = "test.op"() <{prop = memref<*xi4>}> : () -> index
  %3 = "test.op"(%0, %1) : (i6, f32) -> memref<*xindex>
  %4 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i6, i6) -> i6
  "func.return"() : () -> ()
}) : () -> ()
"""

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(RewriteMaybe(recursive=True), apply_recursively=True),
        op_inserted=3,
        op_removed=3,
        op_replaced=3,
        op_modified=1,
    )

    prog = """\
"builtin.module"() ({
  "test.op"() {dict_nest = {hello = i32}} : () -> ()
}) : () -> ()
"""

    expected_recursive = """
"builtin.module"() ({
  "test.op"() {dict_nest = {hello = index}} : () -> ()
}) : () -> ()
"""

    rewrite_and_compare(
        prog,
        prog,
        PatternRewriteWalker(Rewrite(recursive=False)),
        expect_rewrite=False,
    )

    rewrite_and_compare(
        prog,
        expected_recursive,
        PatternRewriteWalker(Rewrite(recursive=True)),
        op_inserted=1,
        op_removed=1,
        op_replaced=1,
    )


def test_recursive_type_conversion_in_regions():
    prog = """\
"builtin.module"() ({
  "func.func"() <{function_type = (memref<2x4xui16>) -> (), sym_name = "main", sym_visibility = "private"}> ({
  ^bb0(%arg0 : memref<2x4xui16>):
    "func.return"() : () -> ()
  }) : () -> ()
}) : () -> ()
"""
    expected_prog = """\
"builtin.module"() ({
  "func.func"() <{function_type = (memref<2x4xindex>) -> (), sym_name = "main", sym_visibility = "private"}> ({
  ^0(%arg0 : memref<2x4xindex>):
    "func.return"() : () -> ()
  }) : () -> ()
}) : () -> ()
"""

    class IndexConversion(TypeConversionPattern):
        @attr_type_rewrite_pattern
        def convert_type(self, typ: IntegerType) -> IndexType:
            return IndexType()

    rewrite_and_compare(
        prog,
        expected_prog,
        PatternRewriteWalker(IndexConversion(recursive=True)),
        op_inserted=1,
        op_removed=1,
        op_replaced=1,
        op_modified=1,
    )


def test_no_change():
    """Test that doing nothing successfully does not report doing something."""

    prog = """\
"builtin.module"() ({
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
}) : () -> ()
"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, matched_op: test.TestOp, rewriter: PatternRewriter):
            return

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
        expect_rewrite=False,
    )


def test_error():
    prog = """\
builtin.module {
  "test.op"() {erroneous = false} : () -> ()
  "test.op"() : () -> ()
  "test.op"() {erroneous = true} : () -> ()
  "test.op"() : () -> ()
}
"""
    expected = """\
Error while applying pattern: Expected operation to not be erroneous!

"builtin.module"() ({
  "test.op"() {erroneous = false} : () -> ()
  "test.op"() : () -> ()
  "test.op"() {erroneous = true} : () -> ()
  ^^^^^^^^^--------------------------------------------------------------
  | Error while applying pattern: Expected operation to not be erroneous!
  -----------------------------------------------------------------------
  "test.op"() : () -> ()
}) : () -> ()

"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, matched_op: test.TestOp, rewriter: PatternRewriter):
            if matched_op.attributes.get(
                "erroneous", IntegerAttr.from_int_and_width(0, 1)
            ) == IntegerAttr.from_int_and_width(1, 1):
                raise ValueError("Expected operation to not be erroneous!")
            return

    ctx = Context(allow_unregistered=True)
    ctx.load_dialect(Builtin)
    ctx.load_dialect(Arith)
    ctx.load_dialect(test.Test)

    parser = Parser(ctx, prog)
    module = parser.parse_module()

    walker = PatternRewriteWalker(Rewrite())
    with pytest.raises(ValueError, match=re.escape(expected)):
        walker.rewrite_module(module)


def test_attr_constr_rewrite_pattern():
    prog = """\
"builtin.module"() ({
  "func.func"() <{function_type = (memref<2x4xui16>) -> (), sym_name = "main", sym_visibility = "private"}> ({
  ^bb0(%arg0 : memref<2x4xui16>):
    "func.return"() : () -> ()
  }) : () -> ()
}) : () -> ()
"""
    expected_prog = """\
"builtin.module"() ({
  "func.func"() <{function_type = (memref<2x4xindex>) -> (), sym_name = "main", sym_visibility = "private"}> ({
  ^0(%arg0 : memref<2x4xindex>):
    "func.return"() : () -> ()
  }) : () -> ()
}) : () -> ()
"""

    class IndexConversion(TypeConversionPattern):
        @attr_constr_rewrite_pattern(BaseAttr(IntegerType))
        def convert_type(self, typ: IntegerType) -> IndexType:
            return IndexType()

    rewrite_and_compare(
        prog,
        expected_prog,
        PatternRewriteWalker(IndexConversion(recursive=True)),
        op_inserted=1,
        op_removed=1,
        op_replaced=1,
        op_modified=1,
    )


def test_pattern_rewriter_erase_op_with_region():
    """Test that erasing an operation with a region works correctly."""
    prog = """
"builtin.module"() ({
  "test.op"() ({
    "test.op"() {error_if_matching} : () -> ()
  }): () -> ()
}) : () -> ()"""
    expected = """
"builtin.module"() ({
^0:
}) : () -> ()"""

    class Rewrite(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: test.TestOp, rewriter: PatternRewriter):
            if "error_if_matching" in op.attributes:
                raise Exception("operation that is supposed to be deleted was matched")
            assert not op.attributes
            rewriter.erase_matched_op()

    rewrite_and_compare(
        prog,
        expected,
        PatternRewriteWalker(Rewrite(), apply_recursively=False),
        op_removed=1,
    )



import pytest

from xdsl.backend.riscv import targets


@pytest.mark.parametrize(
    "simple, expanded",
    [
        ("RV32G", "IMAFDZicsr_Zifencei"),
        ("RV32D", "FDZicsr"),
        ("RV32F", "FZicsr"),
        ("RV32IMZam", "IMAZam"),
    ],
)
def test_march_expansion(simple: str, expanded: str):
    """
    Test that march extensions are expanded correctly
    """
    assert targets.MachineArchSpec(simple).spec_string == expanded


def test_abi_compatibility():
    rv32g = targets.MachineArchSpec("RV32G")
    rv64gq = targets.MachineArchSpec("RV64GQ")
    rv64g = targets.MachineArchSpec("RV64G")
    rv32i = targets.MachineArchSpec("RV32I")
    mabis = targets.MAbi

    # RV32G supports 32bit ABIs with 64bit floats
    assert rv32g.supports_mabi(mabis.ILP32D.value)
    assert rv32g.supports_mabi(mabis.ILP32F.value)
    assert rv32g.supports_mabi(mabis.ILP32.value)

    # RV32G supports no 64 bit ABI
    assert not rv32g.supports_mabi(mabis.LP64D.value)
    assert not rv32g.supports_mabi(mabis.LP64F.value)
    assert not rv32g.supports_mabi(mabis.LP64.value)

    # RV64* should not supprto LP32*
    assert not rv64g.supports_mabi(mabis.ILP32D.value)
    assert not rv64g.supports_mabi(mabis.ILP32.value)

    # RV64G should support 64bit ABIs with up to 64 bit FLEN
    assert rv64g.supports_mabi(mabis.LP64.value)
    assert rv64g.supports_mabi(mabis.LP64F.value)
    assert rv64g.supports_mabi(mabis.LP64D.value)
    # RV64G is missing the Q extension and can therefore not support LP64Q
    assert not rv64g.supports_mabi(mabis.LP64Q.value)
    # RV64GQ has the Q extension and therefore supports it
    assert rv64gq.supports_mabi(mabis.LP64Q.value)

    # RV32I should only support ILP32 and no other ABI
    assert rv32i.supports_mabi(mabis.ILP32.value)
    assert not rv32i.supports_mabi(mabis.LP64.value)
    assert not rv32i.supports_mabi(mabis.ILP32F.value)
    assert not rv32i.supports_mabi(mabis.LP64F.value)



import pytest

from xdsl.dialects.arith import ConstantOp
from xdsl.dialects.builtin import i32
from xdsl.ir import Block, Region


def test_block_insert():
    a = ConstantOp.from_int_and_width(1, i32)
    b = ConstantOp.from_int_and_width(2, i32)
    c = ConstantOp.from_int_and_width(3, i32)
    d = ConstantOp.from_int_and_width(4, i32)
    e = ConstantOp.from_int_and_width(5, i32)

    block = Block()

    assert not block.ops

    assert block.first_op is None
    assert block.last_op is None

    assert list(block.ops) == []

    block.add_op(a)

    assert block.ops

    assert block.first_op is a
    assert block.last_op is a

    assert list(block.ops) == [a]

    block.add_op(c)

    assert block.first_op is a
    assert block.last_op is c

    assert a.next_op is c
    assert c.prev_op is a

    assert list(block.ops) == [a, c]

    block.insert_op_after(b, a)

    assert block.first_op is a
    assert block.last_op is c

    assert a.next_op is b
    assert b.prev_op is a

    assert b.next_op is c
    assert c.prev_op is b

    assert list(block.ops) == [a, b, c]

    block.add_op(e)

    assert block.first_op is a
    assert block.last_op is e

    assert c.next_op is e
    assert e.prev_op is c

    assert list(block.ops) == [a, b, c, e]

    block.insert_op_before(d, e)

    assert block.first_op is a
    assert block.last_op is e

    assert c.next_op is d
    assert d.prev_op is c

    assert d.next_op is e
    assert e.prev_op is d

    assert list(block.ops) == [a, b, c, d, e]


def test_region_insert():
    a = Block((ConstantOp.from_int_and_width(1, i32),))
    b = Block((ConstantOp.from_int_and_width(2, i32),))
    c = Block((ConstantOp.from_int_and_width(3, i32),))
    d = Block((ConstantOp.from_int_and_width(4, i32),))
    e = Block((ConstantOp.from_int_and_width(5, i32),))

    region = Region()

    assert not region.blocks

    assert region.first_block is None
    assert region.last_block is None

    assert list(region.blocks) == []

    region.add_block(a)

    assert region.blocks

    assert region.first_block is a
    assert region.last_block is a

    assert list(region.blocks) == [a]

    region.add_block(c)

    assert region.first_block is a
    assert region.last_block is c

    assert a.next_block is c
    assert c.prev_block is a

    assert list(region.blocks) == [a, c]

    region.insert_block_after(b, a)

    assert region.first_block is a
    assert region.last_block is c

    assert a.next_block is b
    assert b.prev_block is a

    assert b.next_block is c
    assert c.prev_block is b

    assert list(region.blocks) == [a, b, c]

    region.add_block(e)

    assert region.first_block is a
    assert region.last_block is e

    assert c.next_block is e
    assert e.prev_block is c

    assert list(region.blocks) == [a, b, c, e]

    region.insert_block_before(d, e)

    assert region.first_block is a
    assert region.last_block is e

    assert c.next_block is d
    assert d.prev_block is c

    assert d.next_block is e
    assert e.prev_block is d

    assert list(region.blocks) == [a, b, c, d, e]


def test_region_indexing():
    a = Block((ConstantOp.from_int_and_width(1, i32),))
    b = Block((ConstantOp.from_int_and_width(2, i32),))
    c = Block((ConstantOp.from_int_and_width(3, i32),))
    d = Block((ConstantOp.from_int_and_width(4, i32),))
    e = Block((ConstantOp.from_int_and_width(5, i32),))

    region = Region((a, b, c, d, e))

    assert region.blocks[0] == a
    assert region.blocks[1] == b
    assert region.blocks[2] == c
    assert region.blocks[3] == d
    assert region.blocks[4] == e

    with pytest.raises(IndexError):
        region.blocks[5]

    assert region.blocks[-1] == e
    assert region.blocks[-2] == d
    assert region.blocks[-3] == c
    assert region.blocks[-4] == b
    assert region.blocks[-5] == a

    with pytest.raises(IndexError):
        region.blocks[-6]



from xdsl.context import Context
from xdsl.dialects.builtin import AffineMapAttr, ArrayAttr, ModuleOp
from xdsl.dialects.gpu import LoopDimMapAttr, ProcessorAttr, ProcessorEnum
from xdsl.dialects.scf import ParallelOp
from xdsl.ir import Operation
from xdsl.ir.affine import AffineMap
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)

MappingAttrName = "mapping"

MapGrid = 0
MapBlock = 1
Sequential = 2

kNumHardwareIds = 3


def getHardwareIdForMapping(level: int, dimension: int) -> ProcessorEnum:
    """
    Computed the hardware id to use for a given mapping level. Will
    assign x,y and z hardware ids for the first 3 dimensions and use
    sequential after.
    """

    if dimension >= kNumHardwareIds or level == Sequential:
        return ProcessorEnum.Sequential
    match level:
        case 0:
            match dimension:
                case 0:
                    return ProcessorEnum.Block_X
                case 1:
                    return ProcessorEnum.Block_Y
                case 2:
                    return ProcessorEnum.Block_Z
                case _:
                    return ProcessorEnum.Sequential
        case 1:
            match dimension:
                case 0:
                    return ProcessorEnum.Thread_X
                case 1:
                    return ProcessorEnum.Thread_Y
                case 2:
                    return ProcessorEnum.Thread_Z
                case _:
                    return ProcessorEnum.Sequential
        case _:
            return ProcessorEnum.Sequential


def mapParallelOp(parallelOp: ParallelOp, mappingLevel: int = MapGrid):
    """
    Add mapping information to the given parallel loop. Do not add
    mapping information if the loop already has it. Also, don't
    start a mapping at a nested loop.
    """
    # Do not try to add a mapping to already mapped loops or nested loops.
    anchor: Operation | None = parallelOp.parent_op()
    while (anchor is not None) and (not isinstance(anchor, ParallelOp)):
        anchor = anchor.parent_op()

    if (MappingAttrName in parallelOp.attributes) or (
        (mappingLevel == MapGrid) and anchor is not None
    ):
        return
    attrs = [
        getHardwareIdForMapping(mappingLevel, i)
        for i in range(len(parallelOp.lowerBound))
    ]
    attrs = ArrayAttr(
        [
            LoopDimMapAttr(
                [
                    ProcessorAttr(attr),
                    AffineMapAttr(AffineMap.identity(1)),
                    AffineMapAttr(AffineMap.identity(1)),
                ]
            )
            for attr in reversed(attrs)
        ]
    )
    parallelOp.attributes[MappingAttrName] = attrs
    mappingLevel += 1
    for op in parallelOp.body.ops:
        if isinstance(op, ParallelOp):
            mapParallelOp(op, mappingLevel)


class GpuMapParallelLoopsPattern(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ParallelOp, rewriter: PatternRewriter, /):
        mapParallelOp(op)


class GpuMapParallelLoopsPass(ModulePass):
    name = "gpu-map-parallel-loops"

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        walker = PatternRewriteWalker(
            GreedyRewritePatternApplier([GpuMapParallelLoopsPattern()])
        )
        walker.rewrite_module(op)
